# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Authors:
#   Quan Zhou <quan@bitergia.com>
#

import configparser
import logging
import unittest

from grimoire_elk.elk import anonymize_params, enrich_backend, feed_backend, logger


CONFIG_FILE = 'tests.conf'


class TestElk(unittest.TestCase):
    """Unit tests for elk class"""

    @classmethod
    def setUpClass(cls):
        cls.config = configparser.ConfigParser()
        cls.config.read(CONFIG_FILE)
        cls.es_con = dict(cls.config.items('ElasticSearch'))['url']

    def test_anonymize_params(self):
        """Test anonymize parameters"""

        expected_params = ("repo", "--api-token", "xxxxx", "xxxxx", "--sleep-for-rate")
        params = ("repo", "--api-token", "token1", "token2", "--sleep-for-rate")
        anonymized_params = anonymize_params(params)
        self.assertTupleEqual(anonymized_params, expected_params)

        expected_params = ("--backend-password", "xxxxx", "--no-archive", "--api-token", "xxxxx")
        params = ("--backend-password", "mypassword", "--no-archive", "--api-token", "token")
        anonymized_params = anonymize_params(params)
        self.assertTupleEqual(anonymized_params, expected_params)

    def test_feed_backend_wrong_params(self):
        """Test feed_backend with wrong parameters."""

        # Jira has not --api-token argument
        backend_params = ['https://jira.example.org', '--api-token', 'mypersonaltoken']
        projects_json_repo = 'https://jira.example.org'
        expected_msg = "Error feeding raw. Wrong jira arguments: ('https://jira.example.org', " \
                       "'--api-token', 'xxxxx')"
        error_msg = feed_backend(self.es_con, False, False, "jira", backend_params,
                                 "jira_raw", "jira_enriched", projects_json_repo=projects_json_repo)
        self.assertEqual(error_msg, expected_msg)

        # Github the repository name cannot start with '-'
        backend_params = ['chaoss', '-grimoirelab-elk', '--api-token', 'mypersonaltoken']
        projects_json_repo = 'https://github.com/chaoss/-grimoirelab-elk'
        expected_msg = "Error feeding raw. Wrong github arguments: ('chaoss', " \
                       "'-grimoirelab-elk', '--api-token', 'xxxxx')"
        error_msg = feed_backend(self.es_con, False, False, "github", backend_params,
                                 "github_raw", "github_enriched", projects_json_repo=projects_json_repo)
        self.assertEqual(error_msg, expected_msg)

    def test_enrich_backend_wrong_params(self):
        """Test enrich_backend with wrong parameters."""

        # Jira has not --api-token argument
        backend_params = ['https://jira.example.org', '--api-token', 'mypersonaltoken']
        projects_json_repo = 'https://jira.example.org'
        expected_msg = "Error enriching raw. Wrong jira arguments: ('https://jira.example.org', '--api-token', 'xxxxx')"
        with self.assertLogs(logger, level='ERROR') as cm:
            enrich_backend(self.es_con, False, "jira", backend_params, "jira",
                           "jira_raw", "jira_enriched", projects_json_repo=projects_json_repo)
            self.assertEqual(cm.records[0].msg, expected_msg)

        # Github the repository name cannot start with '-'
        backend_params = ['chaoss', '-grimoirelab-elk', '--api-token', 'mypersonaltoken']
        projects_json_repo = 'https://github.com/chaoss/-grimoirelab-elk'
        expected_msg = "Error enriching raw. Wrong github arguments: ('chaoss', " \
                       "'-grimoirelab-elk', '--api-token', 'xxxxx')"
        with self.assertLogs(logger, level='ERROR') as cm:
            enrich_backend(self.es_con, False, "github", backend_params, "github",
                           "github_raw", "github_enriched", projects_json_repo=projects_json_repo)
            self.assertEqual(cm.records[0].msg, expected_msg)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Valerio Cosentino <valcos@bitergia.com>
#

import logging

from .enrich import (Enrich,
                     metadata)
from .utils import fix_field_date
from ..elastic_mapping import Mapping as BaseMapping
from perceval import backend

MAX_SIZE_BULK_ENRICHED_ITEMS = 200

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        Ensure data.message is string, since it can be very large

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
         {
            "dynamic":true,
            "properties": {
                "id": {
                    "type": "keyword"
                },
                "commit_sha": {
                    "type": "keyword"
                },
                "file_path" : {
                    "type" : "keyword"
                },
                "smell" : {
                    "type" : "keyword"
                },
                "origin" : {
                    "type" : "keyword"
                }
            }
        }
        '''

        return {"items": mapping}


class Dockersmells(Enrich):

    def __init__(self, db_sortinghat=None, db_projects_map=None, json_projects_map=None,
                 db_user='', db_password='', db_host=''):
        super().__init__(db_sortinghat, db_projects_map, json_projects_map,
                         db_user, db_password, db_host)

        self.studies = []

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        return identities

    def has_identities(self):
        """ Return whether the enriched items contains identities """

        return False

    def get_field_unique_id(self):
        return "id"

    @metadata
    def get_rich_item(self, item, file_path, smell):
        commit = item['data']
        smell_info = smell.split()
        smell_line = smell_info[1]
        smell_code = smell_info[2]
        smell_details = ' '.join(smell_info[3:])
        eitem = {
            'file_path': file_path,
            'smell_line': smell_line,
            'smell_code': smell_code,
            'smell': smell_details,
            'commit_sha': commit['commit'],
            'origin': item['origin'],
            'author_date': fix_field_date(commit['AuthorDate']),
            'commit_date': fix_field_date(commit['CommitDate'])
        }

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        # uuid
        uuid_sha1 = backend.uuid(eitem['file_path'], eitem['smell'])
        eitem['id'] = "{}_{}".format(eitem['commit_sha'], uuid_sha1)
        eitem.update(self.get_grimoire_fields(eitem["author_date"], "file"))
        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)

        return eitem

    def enrich_items(self, ocean_backend, events=False):
        items_to_enrich = []
        num_items = 0
        ins_items = 0

        for item in ocean_backend.fetch():
            analysis_data = item['data']['analysis']
            for file_path in analysis_data:
                for smell in analysis_data[file_path]['smells']:
                    eitem = self.get_rich_item(item, file_path, smell)
                    items_to_enrich.append(eitem)

            if len(items_to_enrich) < MAX_SIZE_BULK_ENRICHED_ITEMS:
                continue

            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())
            items_to_enrich = []

        if len(items_to_enrich) > 0:
            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())

        if num_items != ins_items:
            missing = num_items - ins_items
            logger.error("[dockersmells] {}/{} missing items".format(
                         missing, num_items))
        else:
            logger.info("[dockersmells] {} items inserted".format(
                        num_items))

        return num_items



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jose Javier Merchante Picazo <jjmerchante@gcauldron.io>
#

from grimoire_elk.identities.identities import Identities


class GitlabIdentities(Identities):
    @classmethod
    def anonymize_item(cls, item):
        """Remove or hash the fields that contain personal information"""

        category = item['category']

        item = item['data']

        if category == "issue":
            identity_types = ['author', 'assignee']
        elif category == "merge_request":
            identity_types = ['author', 'merged_by']
        else:
            identity_types = []

        for identity in identity_types:
            if identity not in item:
                continue
            if not item[identity]:
                continue

            item[identity] = {
                'username': cls._hash(item[identity]['username']),
                'name': cls._hash(item[identity]['name']),
                'email': None,
                'organization': None,
                'location': None
            }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# You should have received a copy of the GNU General Public License
#
# Authors:
#   Valerio Cosentino <valcos@bitergia.com>
#   Nishchith Shetty <inishchith@gmail.com>
#

from grimoirelab_toolkit.datetime import str_to_datetime, unixtime_to_datetime


def get_unique_repository():
    """ Retrieve all the repository names from the index. """

    query_unique_repository = """
    {
        "size": 0,
        "aggs": {
            "unique_repos": {
                "terms": {
                    "field": "origin",
                    "size": 5000
                }
            }
        }
    }
    """

    return query_unique_repository


def get_last_study_date(repository_url, interval):
    """ Retrieve the last study_creation_date of the item corresponding
    to given repository from the study index.
    """

    query_last_study_date = """
    {
        "size": 0,
        "aggs": {
            "1": {
                "max": {
                    "field": "study_creation_date"
                }
            }
        },
        "query": {
            "bool": {
                "filter": [{
                    "term": {
                        "origin.keyword": "%s"
                    }
                },{
                    "term":{
                        "interval_months": "%s"
                    }
                }]
            }
        }
    }
    """ % (repository_url, interval)

    return query_last_study_date


def get_first_enriched_date(repository_url):
    """ Retrieve the first/oldest metadata__updated_on of the item
    corresponding to given repository.
    """

    query_first_enriched_date = """
    {
        "size": 0,
        "aggs": {
            "1": {
                "top_hits": {
                    "docvalue_fields": [
                        "metadata__updated_on"
                    ],
                    "_source": "metadata__updated_on",
                    "size": 1,
                    "sort": [{
                        "commit_date": {
                            "order": "asc"
                        }
                    }]
                }
            }
        },
        "query": {
            "bool": {
                "filter": [{
                    "term": {
                        "origin": "%s"
                    }
                }]
            }
        }
    }
    """ % (repository_url)

    return query_first_enriched_date


def get_files_at_time(repository_url, to_date):
    """ Retrieve all the latest changes wrt files until the to_date,
    corresponding to the given repository.
    """

    query_files_at_time = """
    {
        "size": 0,
        "aggs": {
            "file_stats": {
                "terms": {
                    "field": "file_path",
                    "size": 2147483647,
                    "order": {
                        "_key": "desc"
                    }
                },
                "aggs": {
                    "1": {
                        "top_hits": {
                            "size": 1,
                            "sort": [{
                                "metadata__updated_on": {
                                    "order": "desc"
                                }
                            }]
                        }
                    }
                }
            }
        },
        "query": {
            "bool": {
                "filter": [{
                    "term": {
                        "origin": "%s"
                    }
                },
                {
                    "range": {
                        "metadata__updated_on": {
                            "lte": "%s"
                        }
                    }
                }]
            }
        }
    }
    """ % (repository_url, to_date)

    return query_files_at_time


def get_to_date(es_in, in_index, out_index, repository_url, interval):
    """ Get the appropriate to_date value for incremental insertion. """
    study_data_available = False

    if es_in.indices.exists(index=out_index):
        last_study_date = es_in.search(
            index=out_index,
            body=get_last_study_date(repository_url, interval))["aggregations"]["1"]

        if "value_as_string" in last_study_date and last_study_date["value_as_string"]:
            study_data_available = True
            to_date = str_to_datetime(last_study_date["value_as_string"])
        elif "value" in last_study_date and last_study_date["value"]:
            study_data_available = True
            try:
                to_date = unixtime_to_datetime(last_study_date["value"])
            except Exception:
                to_date = unixtime_to_datetime(last_study_date["value"] / 1000)

    if not study_data_available:
        first_item_date = es_in.search(
            index=in_index,
            body=get_first_enriched_date(repository_url))["aggregations"]["1"]["hits"]["hits"][0]["_source"]

        to_date = str_to_datetime(first_item_date["metadata__updated_on"])

    return to_date



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Nishchith Shetty <inishchith@gmail.com>
#

import logging
from dateutil.relativedelta import relativedelta

from elasticsearch import Elasticsearch as ES, RequestsHttpConnection
from .enrich import (Enrich,
                     metadata)
from .graal_study_evolution import (get_to_date,
                                    get_unique_repository)
from .utils import fix_field_date, anonymize_url
from ..elastic_mapping import Mapping as BaseMapping

from grimoirelab_toolkit.datetime import datetime_utcnow
from grimoire_elk.elastic import ElasticSearch

MAX_SIZE_BULK_ENRICHED_ITEMS = 200

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        Ensure data.message is string, since it can be very large

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
         {
            "dynamic":true,
            "properties": {
                "id" : {
                    "type" : "keyword"
                },
                "interval_months" : {
                    "type" : "long"
                },
                "origin" : {
                    "type" : "keyword"
                },
                "study_creation_date" : {
                    "type" : "date"
                },
                "total_files": {
                    "type": "long"
                },
                "licensed_files": {
                    "type": "long"
                },
                "copyrighted_files": {
                    "type": "long"
                }
            }
        }
        '''

        return {"items": mapping}


class ColicEnrich(Enrich):

    def __init__(self, db_sortinghat=None, db_projects_map=None, json_projects_map=None,
                 db_user='', db_password='', db_host=''):
        super().__init__(db_sortinghat, db_projects_map, json_projects_map,
                         db_user, db_password, db_host)

        self.studies = []
        self.studies.append(self.enrich_colic_analysis)

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        return identities

    def has_identities(self):
        """ Return whether the enriched items contains identities """

        return False

    def get_field_unique_id(self):
        return "id"

    def __get_total_files(self, repository_url, to_date):
        """ Retrieve total number for files until to_date, corresponding
        to the given repository
        """

        query_total_files = """
        {
            "size": 0,
            "aggs": {
                "1": {
                    "cardinality": {
                        "field": "file_path"
                    }
                }
            },
            "query": {
                "bool": {
                    "filter": [{
                        "term": {
                            "origin": "%s"
                        }
                    },
                    {
                        "range": {
                            "metadata__updated_on": {
                                "lte": "%s"
                            }
                        }
                    }]
                }
            }
        }
        """ % (repository_url, to_date)

        return query_total_files

    def __get_licensed_files(self, repository_url, to_date):
        """ Retrieve all the licensed files until the to_date, corresponding
        to the given repository.
        """

        query_licensed_files = """
        {
            "size": 0,
            "aggs": {
                "1": {
                    "cardinality": {
                        "field": "file_path"
                    }
                }
            },
            "query": {
                "bool": {
                    "filter": [{
                        "term": {
                            "has_license": 1
                        }
                    },
                    {
                        "term": {
                            "origin": "%s"
                        }
                    },
                    {
                        "range": {
                            "metadata__updated_on": {
                                "lte": "%s"
                            }
                        }
                    }]
                }
            }
        }
        """ % (repository_url, to_date)

        return query_licensed_files

    def __get_copyrighted_files(self, repository_url, to_date):
        """ Retrieve all the copyrighted files until the to_date, corresponding
        to the given repository.
        """

        query_copyrighted_files = """
        {
            "size": 0,
            "aggs": {
                "1": {
                    "cardinality": {
                        "field": "file_path"
                    }
                }
            },
            "query": {
                "bool": {
                    "filter": [{
                        "term": {
                            "has_copyright": 1
                        }
                    },
                    {
                        "term": {
                            "origin": "%s"
                        }
                    },
                    {
                        "range": {
                            "metadata__updated_on": {
                                "lte": "%s"
                            }
                        }
                    }]
                }
            }
        }
        """ % (repository_url, to_date)

        return query_copyrighted_files

    def extract_modules(self, file_path):
        """ Extracts module path from the given file path """
        path_chunks = file_path.split('/')

        modules = []
        for idx in range(len(path_chunks)):
            sub_path = '/'.join(path_chunks[:idx])

            if sub_path:
                modules.append(sub_path)

        return modules

    @metadata
    def __get_rich_scancode(self, file_analysis):
        # Scancode and Scancode-CLI Implementation

        eitem = {}
        eitem["file_path"] = file_analysis["file_path"]
        eitem["modules"] = self.extract_modules(eitem["file_path"])
        eitem["copyrights"] = []
        eitem["licenses"] = []
        eitem["license_name"] = []
        eitem["has_license"] = 0
        eitem["has_copyright"] = 0

        if file_analysis.get("licenses", False):
            eitem["has_license"] = 1
            for _license in file_analysis["licenses"]:
                eitem["licenses"].extend(_license["matched_rule"]["licenses"])
                eitem["license_name"].append(_license["name"])

        if file_analysis.get("copyrights", False):
            eitem["has_copyright"] = 1
            for _copyright in file_analysis["copyrights"]:
                eitem["copyrights"].append(_copyright["value"])

        return eitem

    @metadata
    def __get_rich_nomossa(self, file_analysis):
        # NOMOS analyzer implementation

        eitem = {}
        eitem["file_path"] = file_analysis["file_path"]
        eitem["modules"] = self.extract_modules(eitem["file_path"])
        eitem["licenses"] = []
        eitem["license_name"] = []
        eitem["has_license"] = 0

        if file_analysis["licenses"] != "No_license_found":
            eitem["has_license"] = 1
            for _license in file_analysis["licenses"]:
                eitem["licenses"].append(_license)
                eitem["license_name"].append(_license)

        # NOMOS doesn't provide copyright information.
        eitem["copyrights"] = []
        eitem["has_copyright"] = 0

        return eitem

    def get_rich_items(self, item):
        """
            :category: code_license_scancode_cli(default)
        """

        if item["category"] == "code_license_nomos":
            get_rich_item = self.__get_rich_nomossa
        else:
            get_rich_item = self.__get_rich_scancode

        entry = item['data']
        enriched_items = []

        for file_analysis in entry["analysis"]:
            eitem = get_rich_item(file_analysis)

            self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

            # common attributes
            eitem['author'] = entry['Author']
            eitem['author_date'] = fix_field_date(entry['AuthorDate'])
            eitem["category"] = item["category"]
            eitem['commit'] = entry['commit']
            eitem['committer'] = entry['Commit']
            eitem['commit_date'] = fix_field_date(entry['CommitDate'])
            eitem['commit_sha'] = entry['commit']
            eitem['message'] = entry.get('message', None)

            # Other enrichment
            eitem["repo_url"] = item["origin"]
            if eitem["repo_url"].startswith('http'):
                eitem["repo_url"] = anonymize_url(eitem["repo_url"])

            if self.prjs_map:
                eitem.update(self.get_item_project(eitem))

            # uuid
            eitem['id'] = "{}_{}".format(eitem['commit_sha'], eitem['file_path'])

            eitem.update(self.get_grimoire_fields(entry["AuthorDate"], "file"))

            self.add_repository_labels(eitem)
            self.add_metadata_filter_raw(eitem)

            enriched_items.append(eitem)

        return enriched_items

    def enrich_items(self, ocean_backend, events=False):
        items_to_enrich = []
        num_items = 0
        ins_items = 0

        for item in ocean_backend.fetch():
            rich_items = self.get_rich_items(item)

            items_to_enrich.extend(rich_items)
            if len(items_to_enrich) < MAX_SIZE_BULK_ENRICHED_ITEMS:
                continue

            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())
            items_to_enrich = []

        if len(items_to_enrich) > 0:
            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())

        if num_items != ins_items:
            missing = num_items - ins_items
            logger.error("[colic] {}/{} missing items".format(missing, num_items))
        else:
            logger.info("[colic] {} items inserted".format(num_items))

        return num_items

    def enrich_colic_analysis(self, ocean_backend, enrich_backend, no_incremental=False,
                              out_index="colic_enrich_graal_repo", interval_months=[3],
                              date_field="grimoire_creation_date"):

        logger.info("[colic] study enrich-colic-analysis start")

        es_in = ES([enrich_backend.elastic_url], retry_on_timeout=True, timeout=100,
                   verify_certs=self.elastic.requests.verify, connection_class=RequestsHttpConnection)
        in_index = enrich_backend.elastic.index
        interval_months = list(map(int, interval_months))

        unique_repos = es_in.search(
            index=in_index,
            body=get_unique_repository())

        repositories = [repo['key'] for repo in unique_repos['aggregations']['unique_repos'].get('buckets', [])]

        logger.info("[colic] study enrich-colic-analysis {} repositories to process".format(
                    len(repositories)))
        es_out = ElasticSearch(enrich_backend.elastic.url, out_index, mappings=Mapping)
        es_out.add_alias("colic_study")

        current_month = datetime_utcnow().replace(day=1, hour=0, minute=0, second=0)
        num_items = 0
        ins_items = 0

        for repository_url in repositories:
            repository_url_anonymized = repository_url
            if repository_url_anonymized.startswith('http'):
                repository_url_anonymized = anonymize_url(repository_url_anonymized)

            logger.info("[colic] study enrich-colic-analysis start analysis for {}".format(
                        repository_url_anonymized))
            evolution_items = []

            for interval in interval_months:

                to_month = get_to_date(es_in, in_index, out_index, repository_url, interval)
                to_month = to_month.replace(month=int(interval), day=1, hour=0, minute=0, second=0)

                while to_month < current_month:
                    copyrighted_files_at_time = es_in.search(
                        index=in_index,
                        body=self.__get_copyrighted_files(repository_url, to_month.isoformat()))

                    licensed_files_at_time = es_in.search(
                        index=in_index,
                        body=self.__get_licensed_files(repository_url, to_month.isoformat()))

                    files_at_time = es_in.search(
                        index=in_index,
                        body=self.__get_total_files(repository_url, to_month.isoformat()))

                    licensed_files = int(licensed_files_at_time["aggregations"]["1"]["value"])
                    copyrighted_files = int(copyrighted_files_at_time["aggregations"]["1"]["value"])
                    total_files = int(files_at_time["aggregations"]["1"]["value"])

                    if not total_files:
                        to_month = to_month + relativedelta(months=+interval)
                        continue

                    evolution_item = {
                        "id": "{}_{}_{}".format(to_month.isoformat(), hash(repository_url_anonymized), interval),
                        "repo_url": repository_url_anonymized,
                        "origin": repository_url,
                        "interval_months": interval,
                        "study_creation_date": to_month.isoformat(),
                        "licensed_files": licensed_files,
                        "copyrighted_files": copyrighted_files,
                        "total_files": total_files
                    }

                    evolution_item.update(self.get_grimoire_fields(evolution_item["study_creation_date"], "stats"))
                    evolution_items.append(evolution_item)

                    if len(evolution_items) >= self.elastic.max_items_bulk:
                        num_items += len(evolution_items)
                        ins_items += es_out.bulk_upload(evolution_items, self.get_field_unique_id())
                        evolution_items = []

                    to_month = to_month + relativedelta(months=+interval)

                if len(evolution_items) > 0:
                    num_items += len(evolution_items)
                    ins_items += es_out.bulk_upload(evolution_items, self.get_field_unique_id())

                if num_items != ins_items:
                    missing = num_items - ins_items
                    logger.error(
                        "[colic] study enrich-colic-analysis {}/{} missing items for Graal CoLic Analysis "
                        "Study".format(missing, num_items)
                    )
                else:
                    logger.info(
                        "[colic] study enrich-colic-analysis {} items inserted for Graal CoLic Analysis "
                        "Study".format(num_items)
                    )

            logger.info(
                "[colic] study enrich-colic-analysis end analysis for {} with month interval".format(
                    repository_url_anonymized)
            )

        logger.info("[colic] study enrich-colic-analysis end")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from .elastic import ElasticOcean
from ..elastic_mapping import Mapping as BaseMapping


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        Non dynamic discovery of type for:
            * data.runs
            * data.actions

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
         {
            "dynamic":true,
                "properties": {
                    "data": {
                        "properties": {
                            "_class" : {
                                "type" : "keyword"
                            },
                            "runs": {
                                "dynamic":false,
                                "properties": {}
                            },
                            "actions": {
                                "dynamic":false,
                                "properties": {}
                            },
                            "building" : {
                                "type" : "boolean"
                            },
                            "builtOn" : {
                                "type" : "keyword"
                            },
                            "changeSet": {
                                "properties": {
                                    "items": {
                                        "properties": {
                                            "comment": {
                                                "type": "text",
                                                "index": true
                                            }
                                        }
                                    }
                                }
                            },
                            "description" : {
                                "type" : "keyword"
                            },
                            "displayName" : {
                                "type" : "keyword"
                            },
                            "duration" : {
                                "type" : "long"
                            },
                            "estimatedDuration" : {
                                "type" : "long"
                            },
                            "executor" : {
                                "type" : "object"
                            },
                            "fullDisplayName" : {
                                "type" : "keyword"
                            },
                            "id" : {
                                "type" : "keyword"
                            },
                              "keepLog" : {
                                "type" : "boolean"
                            },
                              "number" : {
                                "type" : "long"
                            },
                            "queueId" : {
                                "type" : "long"
                            },
                            "result" : {
                                "type" : "keyword"
                            },
                            "subBuilds" : {
                                "type" : "object"
                            },
                            "timestamp" : {
                                "type" : "long"
                            },
                            "url" : {
                                "type" : "keyword"
                            }
                        }
                    }
                }
        }
        '''

        return {"items": mapping}


class JenkinsOcean(ElasticOcean):
    """Jenkins Ocean feeder"""

    mapping = Mapping

    @classmethod
    def get_p2o_params_from_url(cls, url):
        # Jenkins could include in the URL a jenkins-rename-file T1746
        params = {}

        tokens = url.split(' ')  # Just split the URL not the filter
        params['url'] = tokens[0]

        if len(tokens) == 1:
            return params

        for token in tokens[1:]:
            token = token.strip('--')
            assignment = token.split('=')
            fltr_name = assignment[0].strip()
            fltr_value = assignment[1].strip()

            if token.startswith('jenkins'):
                params['jenkins-rename-file'] = fltr_value
            elif token.startswith('filter'):
                params[fltr_name] = fltr_value

        return params

    @classmethod
    def get_perceval_params_from_url(cls, url):
        params = []
        tokens = url.split(' ', 1)  # Just split the URL not jenkins-rename-file
        url = tokens[0]
        params.append(url)

        return params



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from copy import deepcopy

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping
from grimoirelab_toolkit.datetime import str_to_datetime

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "description_analyzed": {
                  "type": "text",
                  "index": true
                }
           }
        } """

        return {"items": mapping}


class CratesEnrich(Enrich):

    mapping = Mapping

    def get_field_author(self):
        return 'owner_user_data'

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        field = self.get_field_author()
        identities.append(self.get_sh_identity(item, field))

        return identities

    def get_sh_identity(self, item, identity_field=None):
        identity = {}

        identity['username'] = None
        identity['email'] = None
        identity['name'] = None

        user = item  # by default a specific user dict is expected
        if isinstance(item, dict) and 'data' in item:
            users = item['data'][identity_field]['users']
            if not users:
                return identity
            else:
                user = users[0]

        if 'login' in user:
            identity['username'] = user['login']
        if 'name' in user:
            identity['name'] = user['name']

        return identity

    def get_field_event_unique_id(self):
        return "download_sample_id"

    def get_rich_events(self, item):
        """
        In the events there are some common fields with the crate. The name
        of the field must be the same in the create and in the downloads event
        so we can filer using it in crate and event at the same time.

        * Fields that don't change: the field does not change with the events
        in a create so the value is always the same in the events of a create.

        * Fields that change: the value of the field changes with events
        """
        if "version_downloads_data" not in item['data']:
            return []

        # To get values from the task
        eitem = self.get_rich_item(item)

        for sample in item['data']["version_downloads_data"]["version_downloads"]:
            event = deepcopy(eitem)
            event['download_sample_id'] = sample['id']
            event['sample_date'] = sample['date']
            sample_date = str_to_datetime(event['sample_date'])
            event['sample_version'] = sample['version']
            event['sample_downloads'] = sample['downloads']
            event.update(self.get_grimoire_fields(sample_date.isoformat(), "downloads_event"))

            yield event

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)
        # The real data
        crate = item['data']

        # data fields to copy
        copy_fields = ["id", "homepage", "name", "repository", "downloads",
                       "description", "recent_downloads", "max_version",
                       "keywords", "categories", "badges", "versions",
                       "updated_at", "created_at"]

        for f in copy_fields:
            if f in crate:
                eitem[f] = crate[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"description": "description_analyzed"}
        for fn in map_fields:
            eitem[map_fields[fn]] = crate[fn]

        # author info
        if crate['owner_user_data']['users']:
            author = crate['owner_user_data']['users'][0]
            eitem['author_id'] = author['id']
            eitem['author_login'] = author['login']
            eitem['author_url'] = author['url']
            eitem['author_avatar'] = author['avatar']

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(eitem['created_at'], "crates"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Jesus M. Gonzalez-Barahona <jgb@bitergia.com>
#

import json
import logging
import os.path
import sys
import unittest

if '..' not in sys.path:
    sys.path.insert(0, '..')

from grimoire_elk.enriched.jira import JiraEnrich


class TestEnrichJira(unittest.TestCase):
    """Functional unit tests for GrimoireELK Enrichment for Jira"""

    def setUp(self):
        self.ritems = []
        self.eitems = []
        with open(os.path.join("data", "jira_raw_fields.json")) as f:
            for line in f:
                self.ritems.append(json.loads(line))
        with open(os.path.join("data", "jira_enriched_fields.json")) as f:
            for line in f:
                self.eitems.append(json.loads(line))

    def test_enrich_fields(self):
        """Test enrich_fields function"""

        for ritem, eitem in zip(self.ritems, self.eitems):
            enriched = {}
            if 'fields' in ritem['_source']['data']:
                JiraEnrich.enrich_fields(ritem['_source']['data']['fields'], enriched)
            self.assertDictEqual(enriched, eitem)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping


logger = logging.getLogger(__name__)


NO_ANCESTOR_TITLE = "NO_TITLE"


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "title_analyzed": {
                  "type": "text",
                  "index": true
                }
           }
        } """

        return {"items": mapping}


class ConfluenceEnrich(Enrich):

    mapping = Mapping

    def get_field_author(self):
        return 'by'

    def get_identities(self, item):
        """ Return the identities from an item """

        field = self.get_field_author()
        yield self.get_sh_identity(item, field)

    def get_project_repository(self, eitem):
        return str(eitem['space'])

    def get_users_data(self, item):
        """ If user fields are inside the global item dict """
        if 'data' in item:
            users_data = item['data']['version']
        else:
            # the item is directly the data (kitsune answer)
            users_data = item
        return users_data

    def get_sh_identity(self, item, identity_field=None):
        identity = {}

        user = item  # by default a specific user dict is expected
        if isinstance(item, dict) and 'data' in item:
            user = item['data']['version'][identity_field]

        identity['username'] = user['username'] if 'username' in user else user.get('publicName', None)
        identity['email'] = user.get('email', None)
        identity['name'] = user.get('displayName', None)

        return identity

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)
        # The real data
        page = item['data']

        # data fields to copy
        copy_fields = ["type", "id", "status", "title", "content_url"]
        for f in copy_fields:
            if f in page:
                eitem[f] = page[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"title": "title_analyzed"}
        for fn in map_fields:
            eitem[map_fields[fn]] = page[fn]

        version = page['version']

        if 'username' in version['by']:
            eitem['author_name'] = version['by']['username']
        else:
            eitem['author_name'] = version['by']['displayName']

        eitem['message'] = None
        if 'message' in version:
            eitem['message'] = version['message']
        eitem['version'] = version['number']
        eitem['date'] = version['when']
        eitem['url'] = page['_links']['base'] + page['_links']['webui']

        if '_expandable' in page and 'space' in page['_expandable']:
            eitem['space'] = page['_expandable']['space']
            eitem['space'] = eitem['space'].replace('/rest/api/space/', '')

        # Ancestors enrichment
        ancestors_titles = []
        ancestors_links = []

        if 'ancestors' in page:
            ancestors = page['ancestors']
            if isinstance(ancestors, list):
                for ancestor in ancestors:
                    if 'title' in ancestor:
                        ancestors_titles.append(ancestor['title'])
                    else:
                        ancestors_titles.append(NO_ANCESTOR_TITLE)

                    ancestors_links.append(ancestor['_links']['webui'])

        eitem['ancestors_titles'] = ancestors_titles
        eitem['ancestors_links'] = ancestors_links

        # Specific enrichment
        if page['type'] == 'page':
            if page['version']['number'] == 1:
                eitem['type'] = 'new_page'
        eitem['is_blogpost'] = 0
        eitem['is_' + eitem['type']] = 1

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(eitem['date'], "confluence"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#
import logging
import time
import unittest
import unittest.mock
from unittest.mock import MagicMock

from base import TestBaseBackend
from grimoire_elk.enriched.enrich import logger, DEMOGRAPHICS_ALIAS
from grimoire_elk.enriched.utils import REPO_LABELS, anonymize_url
from grimoire_elk.raw.github import GitHubOcean


HEADER_JSON = {"Content-Type": "application/json"}


class TestGitHub2(TestBaseBackend):
    """Test GitHub2 backend"""

    connector = "github2"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertEqual(result['items'], 8)
        self.assertEqual(result['raw'], 8)

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()

        self.assertEqual(result['raw'], 7)
        self.assertEqual(result['enrich'], 14)

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['issue_labels'], [])
        self.assertEqual(eitem['reaction_thumb_up'], 0)
        self.assertEqual(eitem['reaction_thumb_down'], 0)
        self.assertEqual(eitem['reaction_confused'], 0)
        self.assertEqual(eitem['reaction_laugh'], 0)
        self.assertEqual(eitem['reaction_total_count'], 0)
        self.assertEqual(item['category'], 'issue')
        self.assertEqual(eitem['user_login'], 'zhquan_example')

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['pull_labels'], ['bug', 'feature'])
        self.assertEqual(item['category'], 'pull_request')
        self.assertNotIn('reaction_thumb_up', eitem)
        self.assertNotIn('reaction_thumb_down', eitem)
        self.assertNotIn('reaction_confused', eitem)
        self.assertNotIn('reaction_laugh', eitem)
        self.assertNotIn('reaction_total_count', eitem)
        self.assertEqual(eitem['time_to_merge_request_response'], 1.0)
        self.assertEqual(eitem['user_login'], 'zhquan_example')

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'repository')
        self.assertEqual(eitem['forks_count'], 16687)
        self.assertEqual(eitem['subscribers_count'], 2904)
        self.assertEqual(eitem['stargazers_count'], 48188)
        self.assertEqual(eitem['url'], "https://github.com/kubernetes/kubernetes")

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'repository')
        self.assertEqual(eitem['forks_count'], 16687)
        self.assertEqual(eitem['subscribers_count'], 4301)
        self.assertEqual(eitem['stargazers_count'], 47118)
        self.assertEqual(eitem['url'], "https://github.com/kubernetes/kubernetes")

        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'repository')
        self.assertEqual(eitem['forks_count'], 1)
        self.assertEqual(eitem['subscribers_count'], 1)
        self.assertEqual(eitem['stargazers_count'], 1)
        self.assertEqual(eitem['url'], "https://github.com/kubernetes/kubernetes")

        item = self.items[5]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'issue')
        self.assertEqual(eitem['user_name'], 'acs')
        self.assertIsNone(eitem['user_domain'])
        self.assertIsNone(eitem['user_org'])
        self.assertEqual(eitem['author_name'], 'acs')
        self.assertIsNone(eitem['assignee_name'])
        self.assertIsNone(eitem['assignee_domain'])
        self.assertIsNone(eitem['assignee_org'])
        self.assertEqual(eitem['reaction_thumb_up'], 0)
        self.assertEqual(eitem['reaction_thumb_down'], 0)
        self.assertEqual(eitem['reaction_confused'], 0)
        self.assertEqual(eitem['reaction_laugh'], 0)
        self.assertEqual(eitem['reaction_total_count'], 0)
        self.assertEqual(eitem['user_login'], 'acs')

        item = self.items[6]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'pull_request')
        self.assertEqual(eitem['user_name'], 'acs')
        self.assertIsNone(eitem['user_domain'])
        self.assertIsNone(eitem['user_org'])
        self.assertEqual(eitem['author_name'], 'acs')
        self.assertIsNone(eitem['merge_author_name'])
        self.assertIsNone(eitem['merge_author_domain'])
        self.assertIsNone(eitem['merge_author_org'])
        self.assertNotIn('reaction_thumb_up', eitem)
        self.assertNotIn('reaction_thumb_down', eitem)
        self.assertNotIn('reaction_confused', eitem)
        self.assertNotIn('reaction_laugh', eitem)
        self.assertNotIn('reaction_total_count', eitem)
        self.assertEqual(eitem['user_login'], 'acs')

        # Check pull requests enriched items
        item = self.items[7]
        eitem = enrich_backend.get_rich_item(item)
        eitems = enrich_backend.enrich_pulls(item, eitem)
        self.assertEqual(len(eitems), 2)

        eitem = eitems[0]
        self.assertEqual(eitem['item_type'], "comment")
        self.assertEqual(eitem['sub_type'], "review_comment")
        self.assertEqual(eitem['body'], "Comment")
        self.assertEqual(eitem['reaction_thumb_up'], 1)
        self.assertEqual(eitem['comment_updated_at'], "2021-12-23T11:13:49Z")
        self.assertEqual(eitem['comment_created_at'], "2021-12-23T10:51:34Z")

        eitem = eitems[1]
        self.assertEqual(eitem['review_state'], "APPROVED")
        self.assertEqual(eitem['item_type'], "comment")
        self.assertEqual(eitem['sub_type'], "review_comment")
        self.assertEqual(eitem['body'], "LGTM")
        self.assertEqual(eitem['comment_updated_at'], "2021-12-27T12:25:49Z")
        self.assertEqual(eitem['comment_created_at'], "2021-12-27T12:25:49Z")

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertGreater(result['raw'], 0)
        self.assertGreater(result['enrich'], 0)
        self.assertGreater(result['enrich'], result['raw'])

        enrich_backend = self.connectors[self.connector][2]()

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_multi_org_names', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        # ... ?

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_refresh_project(self):
        """Test refresh project field for all sources"""

        result = self._test_refresh_project()
        # ... ?

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://github.com/chaoss/grimoirelab-perceval"
        expected_params = [
            'chaoss', 'grimoirelab-perceval'
        ]
        self.assertListEqual(GitHubOcean.get_perceval_params_from_url(url), expected_params)

    def test_geolocation_study(self):
        """ Test that the geolocation study works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_geolocation')

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_geolocation":
                study(ocean_backend, enrich_backend,
                      location_field="user_location", geolocation_field="user_geolocation")

            self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.enrich:[github] Geolocation '
                                           'starting study %s/test_github2_enrich'
                             % anonymize_url(self.es_con))
            self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.enrich:[github] Geolocation '
                                            'end %s/test_github2_enrich'
                             % anonymize_url(self.es_con))

        time.sleep(5)  # HACK: Wait until github enrich index has been written
        items = [item for item in enrich_backend.fetch() if 'user_location' in item]
        self.assertEqual(len(items), 4)
        for item in items:
            self.assertIn('user_geolocation', item)

    def test_feelings_study(self):
        """ Test that feelings study works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_feelings')

        def mocked_feelings(found, nlp_rest_url):
            return '__label__positive', '__label__love'

        enrich_backend.get_feelings = MagicMock(side_effect=mocked_feelings)

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_feelings":
                study(ocean_backend, enrich_backend, attributes=["body"], nlp_rest_url='http://localhost:2901')

            self.assertRegex(cm.output[0], 'INFO:grimoire_elk.enriched.enrich:\\[enrich-feelings\\] Start study.*')
            self.assertRegex(cm.output[-1], 'INFO:grimoire_elk.enriched.enrich:\\[enrich-feelings\\] End study.*')

        time.sleep(5)  # HACK: Wait until github enrich index has been written
        items = [item for item in enrich_backend.fetch() if 'body' in item]
        self.assertEqual(len(items), 7)
        for item in items:
            self.assertEqual(item['has_sentiment'], 1)
            self.assertEqual(item['has_emotion'], 1)
            self.assertEqual(item['feeling_emotion'], '__label__love')
            self.assertEqual(item['feeling_sentiment'], '__label__positive')

    def test_feelings_study_unknown(self):
        """ Test that feelings study works correctly """

        study, ocean_backend, enrich_backend = self._test_study('enrich_feelings')

        def mocked_feelings(found, nlp_rest_url):
            return '__label__unknown', '__label__unknown'

        enrich_backend.get_feelings = MagicMock(side_effect=mocked_feelings)

        with self.assertLogs(logger, level='INFO') as cm:

            if study.__name__ == "enrich_feelings":
                study(ocean_backend, enrich_backend, attributes=["body"], nlp_rest_url='http://localhost:2901')

            self.assertRegex(cm.output[0], 'INFO:grimoire_elk.enriched.enrich:\\[enrich-feelings\\] Start study.*')
            self.assertRegex(cm.output[-1], 'INFO:grimoire_elk.enriched.enrich:\\[enrich-feelings\\] End study.*')

        time.sleep(5)  # HACK: Wait until github enrich index has been written
        items = [item for item in enrich_backend.fetch() if 'body' in item]
        self.assertEqual(len(items), 7)
        for item in items:
            self.assertEqual(item['has_sentiment'], 1)
            self.assertEqual(item['has_emotion'], 1)
            self.assertEqual(item['feeling_emotion'], '__label__unknown')
            self.assertEqual(item['feeling_sentiment'], '__label__unknown')

    def test_demography_study(self):
        """Test that the demography study works correctly"""

        study, ocean_backend, enrich_backend = self._test_study('enrich_demography')

        with self.assertLogs(logger, level='INFO') as cm:
            if study.__name__ == "enrich_demography":
                study(ocean_backend, enrich_backend)

            self.assertEqual(cm.output[0], 'INFO:grimoire_elk.enriched.enrich:[github] Demography '
                                           'starting study %s/test_github2_enrich'
                             % anonymize_url(self.es_con))
            self.assertEqual(cm.output[-1], 'INFO:grimoire_elk.enriched.enrich:[github] Demography '
                                            'end %s/test_github2_enrich'
                             % anonymize_url(self.es_con))

        time.sleep(5)  # HACK: Wait until github2 enrich index has been written
        items = [item for item in enrich_backend.fetch()]
        self.assertEqual(len(items), 14)
        for item in items:
            self.assertNotIn('username:password', item['origin'])
            self.assertNotIn('username:password', item['tag'])
            if 'author_uuid' in item:
                self.assertTrue('demography_min_date' in item.keys())
                self.assertTrue('demography_max_date' in item.keys())

        r = enrich_backend.elastic.requests.get(enrich_backend.elastic.index_url + "/_alias",
                                                headers=HEADER_JSON, verify=False)
        self.assertIn(DEMOGRAPHICS_ALIAS, r.json()[enrich_backend.elastic.index]['aliases'])

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from .elastic import ElasticOcean
from ..elastic_mapping import Mapping as BaseMapping


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
             {
                "dynamic":true,
                    "properties": {
                        "data": {
                            "dynamic":false,
                            "properties": {}
                        },
                        "updated_on": {
                            "type": "long"
                        }
                    }
            }
            '''

        return {"items": mapping}


class NNTPOcean(ElasticOcean):
    """NNTP Ocean feeder"""

    mapping = Mapping

    @classmethod
    def get_perceval_params_from_url(cls, url):
        # In the url the NNTP host and the group are included
        params = url.split()

        return params

    def _fix_item(self, item):
        # Remove all custom fields to avoid the 1000 fields limit in ES

        fields = list(item["data"].keys())
        for field in fields:
            if field.lower().startswith("x-"):
                item["data"].pop(field)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from .utils import get_time_diff_days, anonymize_url
from grimoirelab_toolkit.datetime import (datetime_utcnow, str_to_datetime)

logger = logging.getLogger(__name__)

MAX_LINES_FOR_VOTE = 10


def kafka_kip(enrich):

    """ Kafka Improvement Proposals process study """

    def extract_vote_and_binding(body):
        """ Extracts the vote and binding for a KIP process included in message body """

        vote = 0
        binding = 0  # by default the votes are binding for +1
        nlines = 0

        for line in body.split("\n"):
            if nlines > MAX_LINES_FOR_VOTE:
                # The vote must be in the first MAX_LINES_VOTE
                break
            if line.startswith(">"):
                # This line is from a previous email
                continue
            elif "+1" in line and "-1" in line:
                # Report summary probably
                continue
            elif "to -1" in line or "is -1" in line or "= -1" in line or "-1 or" in line:
                continue
            elif line.startswith("+1") or " +1 " in line or line.endswith("+1") \
                    or " +1." in line or " +1," in line:
                vote = 1
                binding = 1  # by default the votes are binding for +1
                if 'non-binding' in line.lower():
                    binding = 0
                elif 'binding' in line.lower():
                    binding = 1
                break
            elif line.startswith("-1") or line.endswith(" -1") or " -1 " in line \
                    or " -1." in line or " -1," in line:
                vote = -1
                if 'non-binding' in line.lower():
                    binding = 0
                elif 'binding' in line.lower():
                    binding = 1
                break
            nlines += 1

        return (vote, binding)

    def extract_kip(subject):
        """ Extracts a KIP number from an email subject """

        kip = None

        if not subject:
            return kip

        if 'KIP' not in subject:
            return kip

        kip_tokens = subject.split('KIP')
        if len(kip_tokens) > 2:
            # [KIP-DISCUSSION] KIP-7 Security
            for token in kip_tokens:
                kip = extract_kip("KIP" + token)
                if kip:
                    break
            # logger.debug("Several KIPs in %s. Found: %i", subject, kip)
            return kip

        str_with_kip = kip_tokens[1]

        if not str_with_kip:
            # Sample use case subject: Create a space template for KIP
            return kip

        if str_with_kip[0] == '-':
            try:
                # KIP-120: Control
                str_kip = str_with_kip[1:].split(":")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
            try:
                # KIP-8 Add
                str_kip = str_with_kip[1:].split(" ")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
            try:
                # KIP-11- Authorization
                str_kip = str_with_kip[1:].split("-")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
            try:
                # Bound fetch response size (KIP-74)
                str_kip = str_with_kip[1:].split(")")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
            try:
                # KIP-31&
                str_kip = str_with_kip[1:].split("&")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
            try:
                # KIP-31/
                str_kip = str_with_kip[1:].split("/")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
            try:
                # Re: Copycat (KIP-26. PR-99) - plan on moving forward
                str_kip = str_with_kip[1:].split(".")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
        elif str_with_kip[0] == ' ':
            try:
                # KIP 20 Enable
                str_kip = str_with_kip[1:].split(" ")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
            try:
                # Re: [DISCUSS] KIP 88: DescribeGroups Protocol Update
                str_kip = str_with_kip[1:].split(":")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
            try:
                # [jira] [Updated] (KAFKA-5092) KIP 141- ProducerRecordBuilder
                str_kip = str_with_kip[1:].split("-")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass
        elif str_with_kip[0] == ':':
            try:
                # Re: [VOTE] KIP:71 Enable log compaction and deletion to co-exist
                str_kip = str_with_kip[1:].split(" ")[0]
                kip = int(str_kip)
                return kip
            except ValueError:
                pass

        if not kip:
            # logger.debug("Can not extract KIP from %s", subject)
            pass

        return kip

    def lazy_result(votes):
        """ Compute the result of a votation using lazy consensus
            which requires 3 binding +1 votes and no binding vetoes.
        """
        yes = 0
        yes_binding = 0
        veto = 0
        veto_binding = 0

        result = -1

        for (vote, binding) in votes:
            if vote == 1:
                if binding:
                    yes_binding += 1
                else:
                    yes += 1
            if vote == -1:
                if binding:
                    veto_binding += 1
                else:
                    veto += 1

        if veto_binding == 0 and yes_binding >= 3:
            result = 1

        return result

    def add_kip_final_status_field(enrich):
        """ Add kip final status field """

        total = 0

        for eitem in enrich.fetch():
            if "kip" not in eitem:
                # It is not a KIP message
                continue

            if eitem['kip'] in enrich.kips_final_status:
                eitem.update({"kip_final_status":
                              enrich.kips_final_status[eitem['kip']]})
            else:
                logger.warning("[mbox] study Kafka KIP no final status: {}".format(eitem['kip']))
                eitem.update({"kip_final_status": None})
            yield eitem
            total += 1

        logger.info("[mbox] study Kafka KIP total eitems with kafka final status kip field {}".format(total))

    def add_kip_time_status_fields(enrich):
        """ Add kip fields with final status and times """

        total = 0
        max_inactive_days = 90  # days

        enrich.kips_final_status = {}  # final status for each kip

        for eitem in enrich.fetch():
            # kip_status: adopted (closed), discussion (open), voting (open),
            #             inactive (open), discarded (closed)
            # kip_start_end: discuss_start, discuss_end, voting_start, voting_end
            kip_fields = {
                "kip_status": None,
                "kip_discuss_time_days": None,
                "kip_discuss_inactive_days": None,
                "kip_voting_time_days": None,
                "kip_voting_inactive_days": None,
                "kip_is_first_discuss": 0,
                "kip_is_first_vote": 0,
                "kip_is_last_discuss": 0,
                "kip_is_last_vote": 0,
                "kip_result": None,
                "kip_start_end": None
            }

            if "kip" not in eitem:
                # It is not a KIP message
                continue
            kip = eitem["kip"]
            kip_date = str_to_datetime(eitem["email_date"])

            if eitem['kip_is_discuss']:
                kip_fields["kip_discuss_time_days"] = \
                    get_time_diff_days(enrich.kips_dates[kip]['kip_min_discuss'],
                                       enrich.kips_dates[kip]['kip_max_discuss'])

                # Detect first and last discuss messages
                if kip_date == enrich.kips_dates[kip]['kip_min_discuss']:
                    kip_fields['kip_is_first_discuss'] = 1
                    kip_fields['kip_start_end'] = 'discuss_start'
                elif kip_date == enrich.kips_dates[kip]['kip_max_discuss']:
                    kip_fields['kip_is_last_discuss'] = 1
                    kip_fields['kip_start_end'] = 'discuss_end'

                # Detect discussion status
                if "kip_min_vote" not in enrich.kips_dates[kip]:
                    kip_fields['kip_status'] = 'discussion'
                max_discuss_date = enrich.kips_dates[kip]['kip_max_discuss']
                kip_fields['kip_discuss_inactive_days'] = get_time_diff_days(max_discuss_date, datetime_utcnow())

            if eitem['kip_is_vote']:
                kip_fields["kip_voting_time_days"] = \
                    get_time_diff_days(enrich.kips_dates[kip]['kip_min_vote'],
                                       enrich.kips_dates[kip]['kip_max_vote'])

                # Detect first and last discuss messages
                if kip_date == enrich.kips_dates[kip]['kip_min_vote']:
                    kip_fields['kip_is_first_vote'] = 1
                    kip_fields['kip_start_end'] = 'voting_start'
                elif kip_date == enrich.kips_dates[kip]['kip_max_vote']:
                    kip_fields['kip_is_last_vote'] = 1
                    kip_fields['kip_start_end'] = 'voting_end'

                # Detect discussion status
                kip_fields['kip_status'] = 'voting'
                max_vote_date = enrich.kips_dates[kip]['kip_max_vote']
                kip_fields['kip_voting_inactive_days'] = get_time_diff_days(max_vote_date, datetime_utcnow())

                # Now check if there is a result from enrich.kips_scores
                kip_fields['kip_result'] = lazy_result(enrich.kips_scores[kip])

                if kip_fields['kip_result'] == 1:
                    kip_fields['kip_status'] = 'adopted'
                elif kip_fields['kip_result'] == -1:
                    kip_fields['kip_status'] = 'discarded'

            # And now change the status inactive
            if kip_fields['kip_status'] not in ['adopted', 'discarded']:
                inactive_days = kip_fields['kip_discuss_inactive_days']

                if inactive_days and inactive_days > max_inactive_days:
                    kip_fields['kip_status'] = 'inactive'

                inactive_days = kip_fields['kip_voting_inactive_days']
                if inactive_days and inactive_days > max_inactive_days:
                    kip_fields['kip_status'] = 'inactive'

            # The final status is in the kip_is_last_discuss or kip_is_last_vote
            # It will be filled in the next enrichment round
            eitem.update(kip_fields)
            if eitem['kip'] not in enrich.kips_final_status:
                enrich.kips_final_status[kip] = None
            if eitem['kip_is_last_discuss'] and not enrich.kips_final_status[kip]:
                enrich.kips_final_status[kip] = kip_fields['kip_status']
            if eitem['kip_is_last_vote']:
                enrich.kips_final_status[kip] = kip_fields['kip_status']

            yield eitem
            total += 1

        logger.info("[mbox] study Kafka KIP total eitems with kafka extra kip fields {}".format(total))

    def add_kip_fields(enrich):
        """ Add extra fields needed for kip analysis"""

        total = 0

        enrich.kips_dates = {
            0: {
                "kip_min_discuss": None,
                "kip_max_discuss": None,
                "kip_min_vote": None,
                "kip_max_vote": None,
            }
        }

        enrich.kips_scores = {}

        # First iteration
        for eitem in enrich.fetch():
            kip_fields = {
                "kip_is_vote": 0,
                "kip_is_discuss": 0,
                "kip_vote": 0,
                "kip_binding": 0,
                "kip": 0,
                "kip_type": "general"
            }

            kip = extract_kip(eitem['Subject'])
            if not kip:
                # It is not a KIP message
                continue
            if kip not in enrich.kips_dates:
                enrich.kips_dates[kip] = {}
            if kip not in enrich.kips_scores:
                enrich.kips_scores[kip] = []

            kip_date = str_to_datetime(eitem["email_date"])

            # Analyze the subject to fill the kip fields
            if '[discuss]' in eitem['Subject'].lower() or \
               '[kip-discussion]' in eitem['Subject'].lower() or \
               '[discussion]' in eitem['Subject'].lower():
                kip_fields['kip_is_discuss'] = 1
                kip_fields['kip_type'] = "discuss"
                kip_fields['kip'] = kip
                # Update kip discuss dates
                if "kip_min_discuss" not in enrich.kips_dates[kip]:
                    enrich.kips_dates[kip].update({
                        "kip_min_discuss": kip_date,
                        "kip_max_discuss": kip_date
                    })
                else:
                    if enrich.kips_dates[kip]["kip_min_discuss"] >= kip_date:
                        enrich.kips_dates[kip]["kip_min_discuss"] = kip_date
                    if enrich.kips_dates[kip]["kip_max_discuss"] <= kip_date:
                        enrich.kips_dates[kip]["kip_max_discuss"] = kip_date

            if '[vote]' in eitem['Subject'].lower():
                kip_fields['kip_is_vote'] = 1
                kip_fields['kip_type'] = "vote"
                kip_fields['kip'] = kip
                if 'body_extract' in eitem:
                    (vote, binding) = extract_vote_and_binding(eitem['body_extract'])
                    enrich.kips_scores[kip] += [(vote, binding)]
                    kip_fields['kip_vote'] = vote
                    kip_fields['kip_binding'] = binding
                else:
                    logger.debug("[mbox] study Kafka KIP message {} without body".format(eitem['Subject']))
                # Update kip discuss dates
                if "kip_min_vote" not in enrich.kips_dates[kip]:
                    enrich.kips_dates[kip].update({
                        "kip_min_vote": kip_date,
                        "kip_max_vote": kip_date
                    })
                else:
                    if enrich.kips_dates[kip]["kip_min_vote"] >= kip_date:
                        enrich.kips_dates[kip]["kip_min_vote"] = kip_date
                    if enrich.kips_dates[kip]["kip_max_vote"] <= kip_date:
                        enrich.kips_dates[kip]["kip_max_vote"] = kip_date

            eitem.update(kip_fields)
            yield eitem
            total += 1

        logger.info("[mbox] study Kafka KIP total eitems with kafka kip fields {}".format(total))

    logger.debug("[mbox] study Kafka KIP doing from {}".format(anonymize_url(enrich.elastic.index_url)))

    # First iteration with the basic fields
    eitems = add_kip_fields(enrich)
    enrich.elastic.bulk_upload(eitems, enrich.get_field_unique_id())

    # Second iteration with the final time and status fields
    eitems = add_kip_time_status_fields(enrich)
    enrich.elastic.bulk_upload(eitems, enrich.get_field_unique_id())

    # Third iteration to compute the end status field for all KIPs
    eitems = add_kip_final_status_field(enrich)
    enrich.elastic.bulk_upload(eitems, enrich.get_field_unique_id())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alvaro del Castillo <acs@bitergia.com>
#     Valerio Cosentino <valcos@bitergia.com>
#
import logging
import unittest

from base import TestBaseBackend
from grimoire_elk.raw.meetup import MeetupOcean
from grimoire_elk.enriched.utils import REPO_LABELS


class TestMeetup(TestBaseBackend):
    """Test Meetup backend"""

    connector = "meetup"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"
    ocean_index_anonymized = "test_" + connector + "_anonymized"
    enrich_index_anonymized = "test_" + connector + "_enrich_anonymized"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()
        self.assertEqual(result['items'], 3)
        self.assertEqual(result['raw'], 3)

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 3)
        self.assertEqual(result['enrich'], 19)

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['meetup_created'], '2016-03-22T16:36:44+00:00')
        self.assertEqual(eitem['meetup_time'], '2016-04-07T16:30:00+00:00')
        self.assertEqual(eitem['meetup_updated'], '2016-04-07T21:39:24+00:00')
        self.assertEqual(eitem['group_created'], '2016-03-20T15:13:47+00:00')
        self.assertEqual(eitem['group_urlname'], 'sqlpass-es')
        self.assertEqual(eitem['author_uuid'], '029aa3befc96d386e1c7270586f1ec1d673b0b1b')
        self.assertIsNone(eitem['venue_geolocation'])

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['meetup_created'], '2016-05-31T17:30:48+00:00')
        self.assertEqual(eitem['meetup_time'], '2016-06-09T16:45:00+00:00')
        self.assertEqual(eitem['meetup_updated'], '2016-06-09T20:18:18+00:00')
        self.assertEqual(eitem['group_created'], '2016-03-20T15:13:47+00:00')
        self.assertEqual(eitem['group_urlname'], 'sqlpass-es')
        self.assertEqual(eitem['author_uuid'], '810d53ef4a9ae2ebd8064ac690b2e13cfc2df924')
        self.assertIsNotNone(eitem['venue_geolocation'])

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 3)
        self.assertEqual(result['enrich'], 19)

        enrich_backend = self.connectors[self.connector][2]()

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_multi_org_names', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 3)
        self.assertEqual(result['enrich'], 19)

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_refresh_project(self):
        """Test refresh project field for all sources"""

        result = self._test_refresh_project()
        # ... ?

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "South-East-Puppet-User-Group"
        expected_params = [
            "--tag",
            "South-East-Puppet-User-Group",
            "South-East-Puppet-User-Group"
        ]
        self.assertListEqual(MeetupOcean.get_perceval_params_from_url(url), expected_params)

    def test_items_to_raw_anonymized(self):
        """Test whether JSON items are properly inserted into ES anonymized"""

        result = self._test_items_to_raw_anonymized()

        self.assertGreater(result['items'], 0)
        self.assertGreater(result['raw'], 0)
        self.assertEqual(result['items'], result['raw'])

        item = self.items[0]['data']
        self.assertEqual(item['event_hosts'][0]['name'], '3b3e55fdc7886baea165a854d080caf9808cac97')
        self.assertEqual(item['rsvps'][0]['member']['name'], '3b3e55fdc7886baea165a854d080caf9808cac97')
        self.assertEqual(item['rsvps'][1]['member']['name'], '9b0740c20617be08bd6b81a02017e63235cc0204')
        self.assertEqual(item['rsvps'][2]['member']['name'], 'cbd5438b1e1084c1d85bec65a96ca566d9b2ef2e')

        item = self.items[1]['data']
        self.assertEqual(item['event_hosts'][0]['name'], 'aff2cc6caa4228a709ac3bba6b303c7e5dcce550')
        self.assertEqual(item['event_hosts'][1]['name'], '3b3e55fdc7886baea165a854d080caf9808cac97')
        self.assertEqual(item['rsvps'][0]['member']['name'], '3b3e55fdc7886baea165a854d080caf9808cac97')
        self.assertEqual(item['rsvps'][1]['member']['name'], '9b0740c20617be08bd6b81a02017e63235cc0204')
        self.assertEqual(item['rsvps'][2]['member']['name'], 'cbd5438b1e1084c1d85bec65a96ca566d9b2ef2e')
        self.assertEqual(item['comments'][0]['member']['name'], '58668e7669fd564d99db5d581fcdb6a5618440b5')
        self.assertEqual(item['comments'][1]['member']['name'], 'c96634ae1100ab91de991e40bb2fe656bd765de1')

    def test_raw_to_enrich_anonymized(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich_anonymized()

        self.assertEqual(result['raw'], 3)
        self.assertEqual(result['enrich'], 19)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Animesh Kumar <animuz111@gmail.com>
#
import logging
import unittest

import requests
from base import TestBaseBackend
import grimoire_elk.enriched.pagure as pagure_enriched
from grimoire_elk.enriched.utils import REPO_LABELS
from grimoire_elk.raw.pagure import PagureOcean


class TestPagure(TestBaseBackend):
    """Test Pagure backend"""

    connector = "pagure"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertEqual(result['items'], 7)
        self.assertEqual(result['raw'], 7)
        self.assertEqual(result['items'], result['raw'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()

        self.assertEqual(result['raw'], 7)
        self.assertEqual(result['enrich'], 26)

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'issue')
        self.assertEqual(eitem['id'], 1)
        self.assertEqual(eitem['origin'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['tag'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['author_username'], 'animeshk08')
        self.assertEqual(eitem['assignee_username'], 'animeshk08')
        self.assertEqual(eitem['title'], 'Issue Title 1')
        self.assertEqual(eitem['status'], 'Open')
        self.assertIsNone(eitem['close_status'])
        self.assertEqual(eitem['num_comments'], 1)
        self.assertListEqual(eitem['blocks'], ['3'])
        self.assertListEqual(eitem['tags'], [])
        self.assertListEqual(eitem['custom_fields'], [])
        self.assertIsNone(eitem['time_to_close_days'])
        self.assertGreater(eitem['time_open_days'], 0.0)
        self.assertIsNone(eitem['time_to_first_attention'])

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'issue')
        self.assertEqual(eitem['id'], 2)
        self.assertEqual(eitem['origin'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['tag'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['author_username'], 'animeshk08')
        self.assertIsNone(eitem['assignee_username'])
        self.assertEqual(eitem['title'], 'Issue Title 3')
        self.assertEqual(eitem['status'], 'Open')
        self.assertIsNone(eitem['close_status'])
        self.assertEqual(eitem['num_comments'], 4)
        self.assertListEqual(eitem['blocks'], [])
        self.assertListEqual(eitem['tags'], [])
        self.assertIsNotNone(eitem['custom_fields'])
        self.assertIsNone(eitem['time_to_close_days'])
        self.assertGreater(eitem['time_open_days'], 0.0)
        self.assertIsNone(eitem['time_to_first_attention'])

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'issue')
        self.assertEqual(eitem['id'], 3)
        self.assertEqual(eitem['origin'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['tag'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['author_username'], 'animeshk08')
        self.assertEqual(eitem['assignee_username'], 'animeshk0806')
        self.assertEqual(eitem['title'], 'Sample Title 3')
        self.assertEqual(eitem['status'], 'Open')
        self.assertEqual(eitem['milestone'], 'Milestone1')
        self.assertIsNone(eitem['close_status'])
        self.assertEqual(eitem['num_comments'], 9)
        self.assertListEqual(eitem['blocks'], ['2'])
        self.assertListEqual(eitem['tags'], ['Tag1'])
        self.assertListEqual(eitem['custom_fields'], [])
        self.assertIsNone(eitem['time_to_close_days'])
        self.assertGreater(eitem['time_open_days'], 0.0)
        self.assertEqual(eitem['time_to_first_attention'], 0.0)

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'issue')
        self.assertEqual(eitem['id'], 5)
        self.assertEqual(eitem['origin'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['tag'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['author_username'], 'animeshk08')
        self.assertIsNone(eitem['assignee_username'])
        self.assertEqual(eitem['closed_by_username'], 'animeshk08')
        self.assertEqual(eitem['title'], 'Sample title 4')
        self.assertEqual(eitem['status'], 'Closed')
        self.assertIsNone(eitem['milestone'])
        self.assertIsNone(eitem['close_status'])
        self.assertEqual(eitem['num_comments'], 2)
        self.assertListEqual(eitem['blocks'], [])
        self.assertListEqual(eitem['tags'], [])
        self.assertListEqual(eitem['custom_fields'], [])
        self.assertEqual(eitem['time_to_close_days'], 0.0)
        self.assertEqual(eitem['time_open_days'], 0.0)
        self.assertIsNone(eitem['time_to_first_attention'])

        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['category'], 'issue')
        self.assertEqual(eitem['id'], 6)
        self.assertEqual(eitem['origin'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['tag'], 'https://pagure.io/Test-group/Project-example-namespace')
        self.assertEqual(eitem['author_username'], 'animeshk08')
        self.assertEqual(eitem['assignee_username'], 'animeshk0806')
        self.assertEqual(eitem['closed_by_username'], 'animeshk08')
        self.assertEqual(eitem['title'], 'Sample issue 5')
        self.assertEqual(eitem['status'], 'Closed')
        self.assertEqual(eitem['milestone'], 'Milestone1')
        self.assertEqual(eitem['close_status'], 'Close status 1')
        self.assertEqual(eitem['num_comments'], 3)
        self.assertListEqual(eitem['blocks'], [])
        self.assertListEqual(eitem['tags'], ['Tag1'])
        self.assertListEqual(eitem['custom_fields'], [])
        self.assertEqual(eitem['time_to_close_days'], 0.0)
        self.assertEqual(eitem['time_open_days'], 0.0)
        self.assertIsNone(eitem['time_to_first_attention'])

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            if eitem and 'id' in eitem:
                self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 7)
        self.assertEqual(result['enrich'], 26)

        enrich_backend = self.connectors[self.connector][2]()
        enrich_backend.sortinghat = True

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_multi_org_names', source)

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://pagure.io/Test-group/Project-example-namespace"
        expected_params = [
            'Test-group', 'Project-example-namespace'
        ]
        self.assertListEqual(PagureOcean.get_perceval_params_from_url(url), expected_params)

        url = "https://pagure.io/Project-example"
        expected_params = [
            'Project-example'
        ]
        self.assertListEqual(PagureOcean.get_perceval_params_from_url(url), expected_params)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 7)
        self.assertEqual(result['enrich'], 26)

        res = requests.get(self.es_con + "/" + self.enrich_index + "/_search", verify=False)
        for eitem in res.json()['hits']['hits']:
            self.assertEqual(eitem['_source']['project'], "grimoire")

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_refresh_project(self):
        """Test refresh project field for all sources"""

        result = self._test_refresh_project()
        # ... ?

    def test_max_bulk_item_exceed(self):
        """Test bulk upload of documents when number of documents
         exceeds max bulk item limit"""
        pagure_enriched.MAX_SIZE_BULK_ENRICHED_ITEMS = 2
        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 7)
        self.assertEqual(result['enrich'], 26)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            if 'author_name' in eitem:
                for attribute in enrich_backend.RAW_FIELDS_COPY:
                    if attribute in item:
                        self.assertEqual(item[attribute], eitem[attribute])
                    else:
                        self.assertIsNone(eitem[attribute])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from grimoirelab_toolkit.datetime import (unixtime_to_datetime)

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping

from .utils import get_time_diff_days


TASK_OPEN_STATUS = 'Open'
CLOSED_STATUSES = ['resolved', 'wontfix', 'invalid', 'duplicate', 'spite']


logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "main_description_analyzed": {
                  "type": "text",
                  "index": true
                },
                "assigned_to_roles": {
                  "type": "text",
                  "fielddata": true,
                  "index": true
                },
                "tags_analyzed": {
                  "type": "text",
                  "fielddata": true,
                  "index": true
                }
           }
        } """

        return {"items": mapping}


class PhabricatorEnrich(Enrich):

    mapping = Mapping

    roles = ['authorData', 'ownerData']

    def __init__(self, db_sortinghat=None, db_projects_map=None, json_projects_map=None,
                 db_user='', db_password='', db_host=''):
        super().__init__(db_sortinghat, db_projects_map, json_projects_map, db_user, db_password, db_host)

        self.phab_ids_names = {}  # To convert from phab ids to phab names

    def get_field_event_unique_id(self):
        return "transactionID"

    def get_field_author(self):
        return "authorData"

    def get_identities(self, item):
        """ Return the identities from an item """

        if 'authorData' in item['data']['fields']:
            user = self.get_sh_identity(item['data']['fields']['authorData'])
            yield user

        if 'ownerData' in item['data']['fields']:
            user = self.get_sh_identity(item['data']['fields']['ownerData'])
            yield user

    def get_sh_identity(self, item, identity_field=None):
        identity = {}

        identity['email'] = None
        identity['username'] = None
        identity['name'] = None

        if item is None:
            return identity

        user = item
        if isinstance(item, dict) and 'data' in item:
            user = item['data']['fields'][identity_field]

        if user is None:
            return identity

        if 'userName' in user:
            identity['username'] = user['userName']
        if 'realName' in user:
            identity['name'] = user['realName']

        return identity

    def get_users_data(self, item):
        """ If user fields are inside the global item dict """
        if 'data' in item:
            users_data = item['data']['fields']
        else:
            # the item is directly the data (kitsune answer)
            users_data = item
        return users_data

    def get_rich_events(self, item):
        """
        In the events there are some common fields with the task. The name
        of the field must be the same in the task and in the event
        so we can filer using it in task and event at the same time.

        * Fields that don't change: the field does not change with the events
        in a task so the value is always the same in the events of a task.

        * Fields that change: the value of teh field changes with events
        """
        # To get values from the task
        eitem = self.get_rich_item(item)

        # Fields that don't change never
        task_fields_nochange = ['author_userName', 'creation_date', 'url', 'id', 'bug_id']

        # Follow changes in this fields
        task_fields_change = ['priority_value', 'status', 'assigned_to_userName', 'tags_custom_analyzed']
        task_change = {}
        for f in task_fields_change:
            task_change[f] = None
        task_change['status'] = TASK_OPEN_STATUS
        task_change['tags_custom_analyzed'] = eitem['tags_custom_analyzed']

        # Events are in transactions field (changes in fields)
        transactions = item['data']['transactions']

        if not transactions:
            return []

        for t in transactions:
            event = {}
            # Needed for incremental updates from the item
            event['metadata__updated_on'] = item['metadata__updated_on']
            event['origin'] = item['origin']
            # Real event data
            event['transactionID'] = t['transactionID']
            event['type'] = t['transactionType']
            event['username'] = None
            if 'authorData' in t and 'userName' in t['authorData']:
                event['event_author_name'] = t['authorData']['userName']
            event['update_date'] = unixtime_to_datetime(float(t['dateCreated'])).isoformat()
            event['oldValue'] = ''
            event['newValue'] = ''
            if event['type'] == 'core:edge':
                for val in t['oldValue']:
                    if val in self.phab_ids_names:
                        val = self.phab_ids_names[val]
                    event['oldValue'] += "," + val
                event['oldValue'] = event['oldValue'][1:]  # remove first comma
                for val in t['newValue']:
                    if val in self.phab_ids_names:
                        val = self.phab_ids_names[val]
                    event['newValue'] += "," + val
                event['newValue'] = event['newValue'][1:]  # remove first comma
            elif event['type'] in ['status', 'description', 'priority', 'reassign', 'title', 'space', 'core:create', 'parent']:
                # Convert to str so the field is always a string
                event['oldValue'] = str(t['oldValue'])
                if event['oldValue'] in self.phab_ids_names:
                    event['oldValue'] = self.phab_ids_names[event['oldValue']]
                event['newValue'] = str(t['newValue'])
                if event['newValue'] in self.phab_ids_names:
                    event['newValue'] = self.phab_ids_names[event['newValue']]
            elif event['type'] == 'core:comment':
                event['newValue'] = t['comments']
            elif event['type'] == 'core:subscribers':
                event['newValue'] = ",".join(t['newValue'])
            else:
                # logger.debug("Event type %s old to new value not supported", t['transactionType'])
                pass

            for f in task_fields_nochange:
                # The field name must be the same than in task for filtering
                event[f] = eitem[f]

            # To track history of some fields
            if event['type'] in ['status']:
                task_change['status'] = event['newValue']
            elif event['type'] == 'priority':
                task_change['priority'] = event['newValue']
            elif event['type'] == 'core:edge':
                task_change['tags_custom_analyzed'] = [event['newValue']]
            if event['type'] in ['reassign']:
                # Try to get the userName and not the user id
                if event['newValue'] in self.phab_ids_names:
                    task_change['assigned_to_userName'] = self.phab_ids_names[event['newValue']]
                    event['newValue'] = task_change['assigned_to_userName']
                else:
                    task_change['assigned_to_userName'] = event['newValue']
                if event['oldValue'] in self.phab_ids_names:
                    # Try to get the userName and not the user id
                    event['oldValue'] = self.phab_ids_names[event['oldValue']]

            for f in task_change:
                event[f] = task_change[f]

            yield event

    def __fill_phab_ids(self, item):
        """ Get mappings between phab ids and names """
        for p in item['projects']:
            if p and 'name' in p and 'phid' in p:
                self.phab_ids_names[p['phid']] = p['name']
        if 'authorData' not in item['fields'] or not item['fields']['authorData']:
            return
        self.phab_ids_names[item['fields']['authorData']['phid']] = item['fields']['authorData']['userName']
        if 'ownerData' in item['fields'] and item['fields']['ownerData']:
            self.phab_ids_names[item['fields']['ownerData']['phid']] = item['fields']['ownerData']['userName']
        if 'priority' in item['fields']:
            val = item['fields']['priority']['value']
            self.phab_ids_names[str(val)] = item['fields']['priority']['name']
        for t in item['transactions']:
            if 'authorData' in t and t['authorData'] and 'userName' in t['authorData']:
                self.phab_ids_names[t['authorData']['phid']] = t['authorData']['userName']
            elif t['authorData'] and 'name' in t['authorData']:
                # Herald
                self.phab_ids_names[t['authorData']['phid']] = t['authorData']['name']

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.__fill_phab_ids(item['data'])

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)
        # The real data
        phab_item = item['data']

        # data fields to copy
        copy_fields = ["phid", "id", "type"]
        for f in copy_fields:
            if f in phab_item:
                eitem[f] = phab_item[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {
            "id": "bug_id"
        }
        for f in map_fields:
            if f in phab_item:
                eitem[map_fields[f]] = phab_item[f]
            else:
                eitem[map_fields[f]] = None

        eitem['num_changes'] = len(phab_item['transactions'])

        if 'authorData' in phab_item['fields'] and phab_item['fields']['authorData']:
            # eitem['author_roles'] = ",".join(phab_item['fields']['authorData']['roles'])
            eitem['author_roles'] = phab_item['fields']['authorData']['roles']
            eitem['author_userName'] = phab_item['fields']['authorData']['userName']
            eitem['author_realName'] = phab_item['fields']['authorData']['realName']
        if 'ownerData' in phab_item['fields'] and phab_item['fields']['ownerData']:
            eitem['assigned_to_roles'] = phab_item['fields']['ownerData']['roles']
            eitem['assigned_to_userName'] = phab_item['fields']['ownerData']['userName']
            eitem['assigned_to_realName'] = phab_item['fields']['ownerData']['realName']

        eitem['priority'] = phab_item['fields']['priority']['name']
        eitem['priority_value'] = phab_item['fields']['priority']['value']
        eitem['status'] = phab_item['fields']['status']['name']
        eitem['creation_date'] = unixtime_to_datetime(phab_item['fields']['dateCreated']).isoformat()
        eitem['modification_date'] = unixtime_to_datetime(phab_item['fields']['dateModified']).isoformat()
        eitem['update_date'] = unixtime_to_datetime(item['updated_on']).isoformat()
        # raise
        eitem['main_description'] = phab_item['fields']['name']
        eitem['main_description_analyzed'] = eitem['main_description']
        eitem['url'] = eitem['origin'] + "/T" + str(eitem['bug_id'])

        # Time to assign (time to open -> time to assign)
        eitem['time_to_assign_days'] = None
        # Time to attend (time to assign-> time to first activity from assignee)
        eitem['time_to_attend_days'] = None
        # Time to close (time open -> time last updated for closed tasks)
        eitem['closed_at'] = None

        eitem['changes'] = len(phab_item['transactions'])
        # Number of assignments changes
        eitem['changes_assignment'] = 0
        # Number of assignees in the changes
        eitem['changes_assignee_number'] = 0
        # List the changes assignees
        changes_assignee_list = []
        first_assignee_phid = None
        first_assignee_date = None
        # We need to revert them to go from older to newer
        phab_item['transactions'].reverse()
        for change in phab_item['transactions']:
            change_date = unixtime_to_datetime(float(change['dateCreated'])).isoformat()
            if change["transactionType"] == "reassign":
                if not eitem['time_to_assign_days']:
                    eitem['time_to_assign_days'] = get_time_diff_days(eitem['creation_date'], change_date)
                    first_assignee_phid = change['newValue']
                    first_assignee_date = change_date
                if 'authorData' in change and change['authorData'] and 'userName' in change['authorData'] \
                        and change['authorData']['userName'] not in changes_assignee_list:
                    changes_assignee_list.append(change['authorData']['userName'])
                eitem['changes_assignment'] += 1
            elif change["transactionType"] in ["mergedinto"]:
                eitem['closed_at'] = change_date
                eitem['time_to_close_days'] = get_time_diff_days(eitem['creation_date'], eitem['closed_at'])
            elif change["transactionType"] in ["status"]:
                if change['newValue'] in CLOSED_STATUSES:
                    eitem['closed_at'] = change_date
                    eitem['time_to_close_days'] = get_time_diff_days(eitem['creation_date'], eitem['closed_at'])
            if not eitem['time_to_attend_days'] and first_assignee_phid:
                if 'authorData' in change and change['authorData'] and change['authorData']['phid'] == first_assignee_phid:
                    eitem['time_to_attend_days'] = get_time_diff_days(first_assignee_date, change_date)
        eitem['changes_assignee_number'] = len(changes_assignee_list)
        eitem['changes_assignee_list'] = ','.join(changes_assignee_list)
        eitem['comments'] = 0
        for tr in phab_item['transactions']:
            if tr['comments']:
                eitem['comments'] += 1

        eitem['tags'] = []
        for project in phab_item['projects']:
            if project:
                eitem['tags'].append(project['name'])
        eitem['tags_analyzed'] = eitem['tags']
        eitem['tags_custom_analyzed'] = eitem['tags']

        if self.sortinghat:
            eitem.update(self.get_item_sh(item, self.roles))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(eitem['creation_date'], "task"))

        assigned_to = {}
        for f in eitem.keys():
            if 'ownerData' in f:
                # Copy all ownerData data fields to assigned_to fields
                of = f.split('ownerData')[1]
                assigned_to['assigned_to' + of] = eitem[f]
        eitem.update(assigned_to)

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import re
from .elastic import ElasticOcean
from ..elastic_mapping import Mapping as BaseMapping


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        Non dynamic discovery of type for:
            * data.attachments.ts

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
            {
                "dynamic":true,
                "properties": {
                    "data": {
                        "properties": {
                            "attachments": {
                                "dynamic":false,
                                "properties": {}
                            },
                            "channel_info": {
                                "properties": {
                                    "latest": {
                                        "dynamic": false,
                                        "properties": {}
                                    }
                                }
                            },
                            "props": {
                                "dynamic":false,
                                "properties": {
                                    "meeting_id": {
                                        "type": "text",
                                        "index": true
                                    }
                                }
                            },
                            "root": {
                               "dynamic":false,
                                "properties": {}
                            }
                        }
                    }
                }
            }
            '''

        return {"items": mapping}


class MattermostOcean(ElasticOcean):
    """Mattermost Ocean feeder"""

    mapping = Mapping
    URL_REGEX = re.compile('^'  # Pattern starts from the begining of the string

                           # Both legacy and standard URLs start with the base URL
                           r'(?P<base>https?://[\w\.\-]+\.[\w\-]+)'

                           # Now match one of the types of URLs
                           r'(?:'

                           # Pattern for a standard URL
                           r'(/(?P<team>[a-z\-_\d]+)/channels/(?P<channel>[a-z\-_\d]+)/?)'

                           # OR
                           r'|'

                           # Pattern for a legacy URL
                           r'(/? (?P<channel_id>[a-z\d]{26}))'

                           r')$')  # Pattern ends at the end of the string

    @classmethod
    def get_perceval_params_from_url(cls, url):
        """ Get the perceval params given a mattermost URL

        Mattermost URLs can be passed in two formats:
        - Legacy, which looks like <base_url> <channel_id>
        - Standard, which is just the URL for a specific channel.

        An example Legacy "URL" looks like `https://my.mattermost.host/
        993bte1an3dyjmqdgxsr8jr5kh`

        An example Standard URL looks like
        `https://my.mattermost.host/my_team_name/channels/my_channel_name`
        """
        params = []

        # Match the provided URL against the validator/parser regex
        data = cls.URL_REGEX.match(url.lower())

        # If there was no match, the URL was formatted wrong
        if data is None:
            raise RuntimeError(f"Couldn't parse mattermost URL ({url}), unknown format\n"
                               "URLs must be either a url like "
                               "'https://my.mattermost.host/my_team_name/channels/my_channel_name' "
                               "or just a base URL and a channel ID, such as "
                               "'https://my.mattermost.host/ 993bte1an3dyjmqdgxsr8jr5kh'")

        # The `base` group should be matched for both URLs, and is the 1st parameter
        url = data.group('base')
        params.append(url)

        # Extract the other groups
        team = data.group('team')
        channel = data.group('channel')
        channel_id = data.group('channel_id')

        # Check if this is a Standard URL by the presence of the `team` group
        if team is not None:
            params.append(channel)
            params.append(team)
        else:
            params.append(channel_id)

        return params



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

from .elastic import ElasticOcean
from ..elastic_mapping import Mapping as BaseMapping


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        Non dynamic discovery of type for:
            * data.transaction: has string arrays and dicts arrays
        Specific type for:
            * data.fields.priority.subpriority (float)

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = '''
        {
            "dynamic":true,
            "properties": {
                "data": {
                    "properties": {
                        "transactions": {
                            "dynamic":false,
                            "properties": {}
                        },
                        "fields": {
                            "properties": {
                                "priority" : {
                                    "properties": {
                                        "subpriority" : {"type": "float"}
                                    }
                                },
                                "description": {
                                    "dynamic":false,
                                    "properties": {}
                                },
                                "custom_error_stack": {
                                    "type": "text",
                                    "index": true
                                }
                            }
                        }
                    }
                }
            }
        }
        '''

        return {"items": mapping}


class PhabricatorOcean(ElasticOcean):
    """Phabricator Ocean feeder"""

    mapping = Mapping

    def _fix_item(self, item):
        # fields cannot contain dots in ES 2.2. For consistency reason, this fix is applied also
        # to more recent versions of ES

        for field in list(item["data"]["fields"]):
            if '.' in field:
                undotted_field = field.replace('.', '_')
                item["data"]["fields"][undotted_field] = item["data"]["fields"][field]
                item["data"]['fields'].pop(field)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alvaro del Castillo <acs@bitergia.com>
#     Valerio Cosentino <valcos@bitergia.com>
#
import json
import logging
import unittest

from base import TestBaseBackend
from grimoire_elk.raw.jira import JiraOcean
from grimoire_elk.enriched.utils import REPO_LABELS


class TestJira(TestBaseBackend):
    """Test Jira backend"""

    connector = "jira"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()
        self.assertEqual(result['items'], 5)
        self.assertEqual(result['raw'], 5)

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 5)
        self.assertEqual(result['enrich'], 14)

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['data']['id'], "10010")
        self.assertEqual(eitem['id'], "fcf4a8418030d95844f0825fb5b1c3bdc0a1d942_issue_10010_user_creator")
        self.assertEqual(eitem['number_of_comments'], 0)
        self.assertIn("main_description", eitem)
        self.assertIn("main_description_analyzed", eitem)
        self.assertEqual(eitem['author_type'], 'creator')

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item, author_type='assignee')
        self.assertEqual(item['data']['id'], "10010")
        self.assertEqual(eitem['id'], "fcf4a8418030d95844f0825fb5b1c3bdc0a1d942_issue_10010_user_assignee")
        self.assertEqual(eitem['number_of_comments'], 0)
        self.assertIn("main_description", eitem)
        self.assertIn("main_description_analyzed", eitem)
        self.assertEqual(eitem['author_type'], 'assignee')
        self.assertEqual(eitem['status_category_key'], 'done')
        self.assertEqual(eitem['is_closed'], 1)

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item, author_type='reporter')
        self.assertEqual(item['data']['id'], "10010")
        self.assertEqual(eitem['id'], "fcf4a8418030d95844f0825fb5b1c3bdc0a1d942_issue_10010_user_reporter")
        self.assertEqual(eitem['number_of_comments'], 0)
        self.assertIn("main_description", eitem)
        self.assertIn("main_description_analyzed", eitem)
        self.assertEqual(eitem['author_type'], 'reporter')

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['data']['id'], "10008")
        self.assertEqual(eitem['id'], "d6b4168fd458f910fb9af1df0e9edcfaa188cc67_issue_10008_user_creator")
        self.assertEqual(eitem['number_of_comments'], 0)
        self.assertEqual(eitem['author_type'], 'creator')
        self.assertEqual(eitem['status_category_key'], 'done')
        self.assertEqual(eitem['is_closed'], 1)

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['data']['id'], "10017")
        self.assertEqual(eitem['id'], "305dbd15b1f250cb6941d8b9270af3d3a4405084_issue_10017_user_creator")
        self.assertEqual(eitem['number_of_comments'], 0)
        self.assertEqual(eitem['author_type'], 'creator')
        self.assertEqual(eitem['status_category_key'], 'done')
        self.assertEqual(eitem['is_closed'], 1)

        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(item['data']['id'], "10018")
        self.assertEqual(eitem['id'], "929182e386ddb7d290c0dbd2eb34140993c8f567_issue_10018_user_creator")
        self.assertEqual(eitem['number_of_comments'], 2)
        self.assertEqual(eitem['author_type'], 'creator')
        self.assertEqual(eitem['status_category_key'], 'done')
        self.assertEqual(eitem['is_closed'], 1)

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn(REPO_LABELS, eitem)

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn(REPO_LABELS, eitem)

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn(REPO_LABELS, eitem)

        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 5)
        self.assertEqual(result['enrich'], 14)

        enrich_backend = self.connectors[self.connector][2]()
        enrich_backend.sortinghat = True

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_name'], 'Maurizio Pillitu')
        self.assertEqual(eitem['author_org_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'Unknown')
        self.assertEqual(eitem['author_type'], 'creator')
        self.assertEqual(eitem['author_multi_org_names'], ['Unknown'])

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item, author_type='assignee')
        self.assertEqual(eitem['author_name'], 'Peter Monks')
        self.assertEqual(eitem['author_org_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'peter')
        self.assertEqual(eitem['author_type'], 'assignee')
        self.assertEqual(eitem['author_multi_org_names'], ['Unknown'])

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item, author_type='reporter')
        self.assertEqual(eitem['author_name'], 'Maurizio Pillitu')
        self.assertEqual(eitem['author_org_name'], 'Unknown')
        self.assertEqual(eitem['author_user_name'], 'maoo')
        self.assertEqual(eitem['author_type'], 'reporter')
        self.assertEqual(eitem['author_multi_org_names'], ['Unknown'])

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 5)
        self.assertEqual(result['enrich'], 14)

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_refresh_project(self):
        """Test refresh project field for all sources"""

        result = self._test_refresh_project()
        # ... ?

    def test_get_p2o_params_from_url(self):
        """Test the extraction of p2o params from the projects.json entry"""

        with open("data/projects-release.json") as projects_filename:
            url = json.load(projects_filename)['grimoire']['jira'][0]
            p2o_params = {'url': 'https://jira.opnfv.org', 'filter-raw': 'data.fields.project.key:PROJECT-KEY'}
            self.assertDictEqual(p2o_params, JiraOcean.get_p2o_params_from_url(url))

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        eitems = [enrich_backend.get_rich_item(self.items[0]),
                  enrich_backend.get_rich_item(self.items[0], author_type='assignee'),
                  enrich_backend.get_rich_item(self.items[0], author_type='reporter')]

        for eitem in eitems:
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in self.items[0]:
                    self.assertEqual(self.items[0][attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Alberto Pérez García-Plaza <alpgarcia@bitergia.com>
#

"""Stuff to deal with index patterns from JSON files genrated by Kidash."""
import csv
import logging

from unittest import TestCase


class Schema(object):
    """Intends to represent index structure in order
    to compare ES mappings against other compatible
    structures whose types does not match one to one,
    like Kibana types.

    schema_name     -- schema identifier
    __properties    -- dictionary of properties. Example:
        {
            'properties':{
                'author_name': {
                    'type': 'text',
                    'analyzed':  true,
                    'agg': true
                }
                ...
            }
        }

    """

    # List of types allowed for properties in this Schema
    __supported_types = ['text', 'keyword', 'number', 'date', 'boolean',
                         '_source', 'geo_point', 'object']
    # ES meta fields are excluded using the following list
    __excluded_props = ['_all', '_field_names', '_id', '_index', '_meta',
                        '_parent', '_score', '_routing', '_source', '_type',
                        '_uid']
    # In mbox enriched indexes is_mbox is defined
    __excluded_props += ['is_gmane_message', 'is_mbox_message']
    # Cache from askbot could not include this field
    __excluded_props += ['is_askbot_comment']
    # Cache from confluence could not include this field
    __excluded_props += ['is_attachment', 'is_comment', 'is_new_page']

    def __init__(self, schema_name):
        self.schema_name = schema_name
        self.properties = {}

    @staticmethod
    def check_type(src_type):
        """Checks if type is valid

        TypeError -- when type is not supported
        """
        if src_type not in Schema.__supported_types:
            raise TypeError("Type not supported:", src_type)

        return src_type

    def add_property(self, pname, ptype, agg=False):
        """Adds a new property to the current schema. If there
        is already a property with same name, that property
        will be updated.

        Each schema property will have 3 fields:
          - name
          - type
          - agg

        Keyword arguments:
        pname       -- property names
        ptype       -- property type
        agg         -- is this property aggregatable in Kibana?
        """
        # Exclude ES internal properties
        if pname not in self.__excluded_props:
            schema_type = self.check_type(ptype)
            self.properties[pname] = {'type': schema_type}

            if agg:
                self.properties[pname]['agg'] = True

    def get_properties(self):
        """Returns a dictionary containing all schema properties.

            Example:
            {
                'properties':{
                    'author_name': {
                        'type': 'text',
                        'analyzed':  true,
                        'agg': true
                    }
                    ...
                }
            }
        """
        return self.properties

    def compare_properties(self, schema):
        """Compares two schemas. The schema used to call the method
        will be the one we compare from, in advance 'source schema'.
        The schema passed as parameter will be the 'target schema'.

        Returns -- dictionary {status, correct, missing, distinct, message}
            being:
                status 'OK' if all properties in source schema exist in target
                    schema with same values. 'KO' in other case.
                correct: list of properties that matches.
                missing: list of properties missing from target schema.
                distinct: list of properties in both schemas but having
                    with different values.
                message: a string with additional information.
        """
        test_case = TestCase('__init__')
        test_case.maxDiff = None

        status = 'OK'
        correct = []
        missing = []
        distinct = []
        msg = ''
        for pname, pvalue in self.get_properties().items():
            if pname not in schema.get_properties():
                missing.append(pname)
                msg = msg + '\n' + '* Missing property: ' + pname
                status = 'KO'
            else:
                try:
                    test_case.assertDictEqual(pvalue,
                                              schema.get_properties()[pname])
                    correct.append(pname)

                except AssertionError as e:
                    distinct.append(pname)
                    msg = "%s\n* Type mismatch: \n\t%s: %s" %\
                        (msg, pname, str(e).replace('\n', '\n\t'))
                    status = 'KO'

        return {'status': status, 'correct': correct, 'missing': missing,
                'distinct': distinct, 'msg': msg}


class ESMapping(Schema):
    """Represents an ES Mapping"""

    __mapping_types = {'long': 'number',
                       'integer': 'number',
                       'float': 'number',
                       'double': 'number'}
    __non_aggregatables = {'text'}

    @classmethod
    def from_csv(cls, index_name, csv_file):
        """Builds a ESMapping object from a CSV schema definition

        CSV Format:
        field_name, es_type

        :param raw_csv: CSV file to parse
        :returns: ESMapping
        """
        es_mapping = cls(schema_name=index_name)

        with open(csv_file) as f:
            reader = csv.DictReader(f, delimiter=',')
            for row in reader:
                # With current CSVs is not possible to know which
                # fields should be aggregatable. If we rely on
                # '__non_aggregatables' list, then text fields would
                # be non-aggregatables by default. Nevertheless, we
                # want them to be aggregatables in most cases. As a
                # quick fix while CSV schema specification improves,
                # we will assume that fields defined as text in CSVs
                # must be aggregatables in ESMapping.
                property_type = row['type']
                if property_type == 'text':
                    agg = True
                else:
                    agg = None
                es_mapping.add_property(pname=row['name'], ptype=property_type,
                                        agg=agg)

        return es_mapping

    @classmethod
    def from_json(cls, index_name, mapping_json):
        """Builds a ESMapping object from an ES JSON mapping

        We don't know the exact name due to the aliases:
            GET git/_mapping

        Could retrieve:
            {
                "git_enriched": {
                   "mappings": {
                       "items": {
                           "dynamic_templates": [
                               {...}
                           ],
                           "properties": {
                               "Author": {
                                   ...

        So we need to take name as input parameter.

        We can even find several indices if an alias groups them:
           GET affiliations/_mapping

           {
                "git_enriched": {
                   ...
                },
                "gerrit_enriched": {
                   ...

        We store all properties together in the same Schema.
        """
        es_mapping = cls(schema_name=index_name)

        # Get a list of all index mappings in JSON
        mapping_list = list(mapping_json.values())

        # Extract properties from all those grouped indices
        for nested_json in mapping_list:
            # nested_json = mapping_list[0]
            items = nested_json['mappings']['items']['properties'].items()
            for prop, value in items:
                # Support for nested properties:
                # "channel_purpose": {
                #   "properties": {
                #     "value": {
                #       "type": "keyword"
                #     },
                #     "creator": {
                #       "type": "keyword"
                #     },
                #     "last_set": {
                #       "type": "long"
                #     }
                #   }
                # },
                if 'properties' in value:
                    for nested_prop, nested_value in value['properties'].items():
                        prop_name = prop + '.' + nested_prop
                        agg = None

                        if 'fielddata' in nested_value:
                            agg = nested_value['fielddata']
                        if 'type' in nested_value:
                            ptype = nested_value['type']
                        else:
                            logging.warning('Not adding to es_mapping checking the nested value: %s',
                                            nested_value)
                            continue
                        es_mapping.add_property(pname=prop_name,
                                                ptype=ptype,
                                                agg=agg)

                # Support for "regular" properties
                # "channel_id": {
                #   "type": "keyword"
                # },
                else:
                    agg = None
                    if 'fielddata' in value:
                        agg = value['fielddata']
                    es_mapping.add_property(pname=prop, ptype=value['type'],
                                            agg=agg)

        return es_mapping

    @staticmethod
    def get_schema_type(src_type):
        """Type conversion from ES mapping types to schema types.
        """
        out_type = src_type
        if src_type in ESMapping.__mapping_types:
            out_type = ESMapping.__mapping_types[src_type]

        return out_type

    def add_property(self, pname, ptype, agg=None):
        """Overwrites parent method for type conversion
        """
        schema_type = self.get_schema_type(src_type=ptype)

        # Those fields not explicitely specified as aggregatable or not will be
        # aggregatables depending on their type
        schema_agg = agg
        if schema_agg is None:
            if schema_type in self.__non_aggregatables:
                schema_agg = False
            else:
                schema_agg = True
        super().add_property(pname, schema_type, schema_agg)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Nitish Gupta <imnitish.ng@gmail.com>
#
import logging
import unittest

import requests
from base import TestBaseBackend
from grimoire_elk.enriched.utils import REPO_LABELS
from grimoire_elk.raw.gitter import GitterOcean


class TestGitter(TestBaseBackend):
    """Test Gitter backend"""

    connector = "gitter"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identities(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()

        self.assertEqual(result['items'], 6)
        self.assertEqual(result['raw'], 6)

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 6)
        self.assertEqual(result['enrich'], 6)

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['id'], '1jf932nsdaf122h3292n402y')
        self.assertEqual(eitem['origin'], 'https://gitter.im/test_org/test_room')
        self.assertEqual(eitem['readBy'], 19)
        self.assertEqual(eitem['repository_labels'], None)
        self.assertEqual(eitem['tag'], 'https://gitter.im/test_org/test_room')
        self.assertIn('text_analyzed', eitem)
        self.assertEqual(eitem['unread'], 0)
        self.assertEqual(eitem['uuid'], '1a8b34f51ac1d831d095c6e39fae0a7e600cfcc7')

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['id'], '5e90c74af450c25cc8d05649')
        self.assertEqual(eitem['is_gitter_message'], 1)
        self.assertEqual(eitem['origin'], 'https://gitter.im/test_org/test_room')
        self.assertEqual(eitem['readBy'], 17)
        self.assertEqual(eitem['repository_labels'], None)
        self.assertEqual(eitem['tag'], 'https://gitter.im/test_org/test_room')
        self.assertIn('text_analyzed', eitem)
        self.assertIn('issues', eitem)
        self.assertEqual(eitem['issues'][0]['is_issue'], 'jenkinsci/docker #939')
        self.assertEqual(eitem['issues'][0]['url'], 'https://github.com/jenkinsci/docker/issues/939')
        self.assertEqual(eitem['issues'][0]['repo'], 'jenkinsci/docker')
        self.assertEqual(eitem['unread'], 0)
        self.assertEqual(eitem['uuid'], 'c42df902ac8277c70afb8296af6bd09b5ffa31d6')

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['id'], '5e879554f1207e2adc0cfe53')
        self.assertEqual(eitem['is_gitter_message'], 1)
        self.assertEqual(eitem['origin'], 'https://gitter.im/test_org/test_room')
        self.assertEqual(eitem['readBy'], 34)
        self.assertEqual(eitem['repository_labels'], None)
        self.assertEqual(eitem['tag'], 'https://gitter.im/test_org/test_room')
        self.assertIn('text_analyzed', eitem)
        self.assertIn('issues', eitem)
        self.assertEqual(eitem['issues'][0]['is_pull'], 'jenkinsci/jenkins #3861')
        self.assertEqual(eitem['issues'][0]['url'], 'https://github.com/jenkinsci/jenkins/pull/3861')
        self.assertEqual(eitem['issues'][0]['repo'], 'jenkinsci/jenkins')
        self.assertEqual(eitem['unread'], 0)
        self.assertEqual(eitem['uuid'], 'e8266108628134ae9ba2b1cf08ea3dffefebf8e3')

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['id'], '5e87070981a582042e974777')
        self.assertEqual(eitem['is_gitter_message'], 1)
        self.assertEqual(eitem['origin'], 'https://gitter.im/test_org/test_room')
        self.assertEqual(eitem['readBy'], 28)
        self.assertEqual(eitem['repository_labels'], None)
        self.assertEqual(eitem['tag'], 'https://gitter.im/test_org/test_room')
        self.assertIn('text_analyzed', eitem)
        self.assertIn('mentioned', eitem)
        self.assertEqual(eitem['mentioned'][0]['mentioned_username'], 'IbrahimPatel89')
        self.assertEqual(eitem['mentioned'][0]['mentioned_userId'], '5e870448d73408ce4fdf18b6')
        self.assertEqual(eitem['unread'], 0)
        self.assertEqual(eitem['uuid'], '42ca765de9425d85ac2aebb6750701b3c5f85dcc')

        item = self.items[4]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['id'], '5e85efad1742080c5dfffab0')
        self.assertEqual(eitem['is_gitter_message'], 1)
        self.assertEqual(eitem['origin'], 'https://gitter.im/test_org/test_room')
        self.assertEqual(eitem['readBy'], 24)
        self.assertEqual(eitem['repository_labels'], None)
        self.assertEqual(eitem['tag'], 'https://gitter.im/test_org/test_room')
        self.assertIn('text_analyzed', eitem)
        self.assertIn('url_hostname', eitem)
        self.assertEqual(eitem['url_hostname'][0], 'https://github.com/')
        self.assertEqual(eitem['unread'], 1)
        self.assertEqual(eitem['uuid'], '2bdae2762b5a132669869048a0f07c0bef7f15cb')

        item = self.items[5]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['id'], '5e85efad1742650c5dfffab0')
        self.assertEqual(eitem['is_gitter_message'], 1)
        self.assertEqual(eitem['origin'], 'https://gitter.im/test_org/test_room')
        self.assertEqual(eitem['readBy'], 1)
        self.assertEqual(eitem['repository_labels'], None)
        self.assertEqual(eitem['tag'], 'https://gitter.im/test_org/test_room')
        self.assertIn('text_analyzed', eitem)
        self.assertIn('url_hostname', eitem)
        self.assertIn('mentioned', eitem)
        self.assertIn('issues', eitem)
        self.assertEqual(eitem['unread'], 1)
        self.assertEqual(eitem['uuid'], '2bdae2762b5a132669869048a0f07c0bef777667')
        self.assertEqual(eitem['url_hostname'][1], 'https://facebook.com/')
        self.assertEqual(eitem['issues'][1]['is_pull'], 'jenkinsci/jenkins #3863')
        self.assertEqual(eitem['mentioned'][1]['mentioned_username'], 'devanshu123')

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 6)
        self.assertEqual(result['enrich'], 6)

        enrich_backend = self.connectors[self.connector][2]()
        enrich_backend.sortinghat = True

        url = self.es_con + "/" + self.enrich_index + "/_search"
        response = enrich_backend.requests.get(url, verify=False).json()
        for hit in response['hits']['hits']:
            source = hit['_source']
            if 'author_uuid' in source:
                self.assertIn('author_domain', source)
                self.assertIn('author_gender', source)
                self.assertIn('author_gender_acc', source)
                self.assertIn('author_org_name', source)
                self.assertIn('author_bot', source)
                self.assertIn('author_name', source)
                self.assertIn('author_multi_org_names', source)
                self.assertIn('fromUser_gender', source)
                self.assertIn('fromUser_gender_acc', source)
                self.assertIn('fromUser_multi_org_names', source)
                self.assertIn('fromUser_name', source)
                self.assertIn('fromUser_user_name', source)
                self.assertIn('fromUser_uuid', source)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 6)
        self.assertEqual(result['enrich'], 6)

        res = requests.get(self.es_con + "/" + self.enrich_index + "/_search", verify=False)
        for eitem in res.json()['hits']['hits']:
            self.assertEqual(eitem['_source']['project'], "grimoire")

    def test_refresh_identities(self):
        """Test refresh identities"""

        result = self._test_refresh_identities()
        # ... ?

    def test_refresh_project(self):
        """Test refresh project field for all sources"""

        result = self._test_refresh_project()
        # ... ?

    def test_perceval_params(self):
        """Test the extraction of perceval params from an URL"""

        url = "https://gitter.im/test_org/test_room"
        expected_params = [
            'test_org', 'test_room'
        ]
        self.assertListEqual(GitterOcean.get_perceval_params_from_url(url), expected_params)

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import json
import logging

from grimoire_elk.enriched.enrich import Enrich, metadata, anonymize_url
from ..elastic_mapping import Mapping as BaseMapping


logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "Event_Description": {
                    "type": "keyword"
                },
                "Event_Description_Analyzed": {
                    "type": "text",
                    "index": true
                },
                "Event_Creations": {
                    "type": "keyword"
                },
                "Event_Creations_Analyzed": {
                    "type": "text",
                    "index": true
                },
                "Feedback_from_Attendees": {
                    "type": "keyword"
                },
                "Feedback_from_Attendees_Analyzed": {
                    "type": "text",
                    "index": true
                },
                "Your_Feedback": {
                    "type": "keyword"
                },
                "Your_Feedback_Analyzed": {
                    "type": "text",
                    "index": true
                },
                "geolocation": {
                    "type": "geo_point"
                }
           }
        } """

        return {"items": mapping}


class MozillaClubEnrich(Enrich):

    mapping = Mapping

    def set_elastic(self, elastic):
        self.elastic = elastic

    def get_field_author(self):
        return "Your Name"

    def get_field_date(self):
        return "metadata__updated_on"

    def get_field_unique_id(self):
        return "uuid"

    def get_identities(self, item):
        """Return the identities from an item"""

        user = self.get_sh_identity(item, self.get_field_author())
        yield user

    def get_sh_identity(self, item, identity_field):
        identity = {}

        identity['username'] = item['data'][identity_field]
        identity['email'] = None
        identity['name'] = item['data'][identity_field]

        return identity

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        # metadata fields to copy
        copy_fields = ["metadata__updated_on", "metadata__timestamp",
                       "origin", "uuid"]
        for f in copy_fields:
            if f in item:
                eitem[f] = item[f]
            else:
                eitem[f] = None
        # The real data
        event = item['data']

        if "Event Description" in event and event["Event Description"]:
            event["Event Description"] = event["Event Description"][:self.KEYWORD_MAX_LENGTH]
            event["Event Description Analyzed"] = event["Event Description"]

        if "Event Creations" in event and event["Event Creations"]:
            event["Event Creations"] = event["Event Creations"][:self.KEYWORD_MAX_LENGTH]
            event["Event Creations Analyzed"] = event["Event Creations"]

        if "Feedback from Attendees" in event and event["Feedback from Attendees"]:
            event["Feedback from Attendees"] = event["Feedback from Attendees"][:self.KEYWORD_MAX_LENGTH]
            event["Feedback from Attendees Analyzed"] = event["Feedback from Attendees"]

        if "Your Feedback" in event and event["Your Feedback"]:
            event["Your Feedback"] = event["Your Feedback"][:self.KEYWORD_MAX_LENGTH]
            event["Your Feedback Analyzed"] = event["Your Feedback"]

        # just copy all fields converting in field names spaces to _
        for f in event:
            eitem[f.replace(" ", "_")] = event[f]

        # Use the unique id from perceval
        eitem['id'] = eitem['uuid']
        # There is no url
        eitem['url'] = None

        # Transform numeric fields
        eitem['Attendance'] = int(eitem['Attendance'])

        if self.sortinghat:
            eitem.update(self.get_item_sh(item, "Your Name"))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(event["Timestamp"], "event"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def enrich_items(self, ocean_backend):
        max_items = self.elastic.max_items_bulk
        current = 0
        total = 0
        bulk_json = ""

        url = self.elastic.get_bulk_url()

        logger.debug("[mozillaclub] Adding items to {} (in {} packs)".format(anonymize_url(url), max_items))

        items = ocean_backend.fetch()
        for item in items:
            if current >= max_items:
                total += self.elastic.safe_put_bulk(url, bulk_json)
                bulk_json = ""
                current = 0

            rich_item = self.get_rich_item(item)
            data_json = json.dumps(rich_item)
            bulk_json += '{"index" : {"_id" : "%s" } }\n' % \
                (item[self.get_field_unique_id()])
            bulk_json += data_json + "\n"  # Bulk document
            current += 1

        if current > 0:
            total += self.elastic.safe_put_bulk(url, bulk_json)

        return total
